#include "samda1e16b.h"
#include <stdbool.h>
#include <string.h>
#include <stdint.h>
#include <stdio.h>
#include "printf.h"
#include "sd.h"



#define FLAG_TIMEOUT		0x80
#define FLAG_PARAM_ERR		0x40
#define FLAG_ADDR_ERR		0x20
#define FLAG_ERZ_SEQ_ERR	0x10
#define FLAG_CMD_CRC_ERR	0x08
#define FLAG_ILLEGAL_CMD	0x04
#define FLAG_ERZ_RST		0x02
#define FLAG_IN_IDLE_MODE	0x01


struct SD {
	uint8_t HC		:1;
	uint8_t SD		:1;
};


#define VERBOSE					0


static struct SD mSD;

static void sdHwPrvSetBrg(uint_fast8_t brg)
{
	SERCOM3->SPI.CTRLA.bit.ENABLE = 0;
	while(SERCOM3->SPI.SYNCBUSY.bit.ENABLE);
	SERCOM3->SPI.BAUD.reg = brg;
	SERCOM3->SPI.CTRLA.bit.ENABLE = 1;
	while(SERCOM3->SPI.SYNCBUSY.bit.ENABLE);
}

static bool sdHwInit(void)
{
	uint32_t ctrla;
	
	SERCOM3->SPI.CTRLA.reg = 0;
	while (SERCOM3->SPI.SYNCBUSY.bit.ENABLE);
	SERCOM3->SPI.CTRLB.bit.RXEN = 1;
	SERCOM3->SPI.BAUD.reg = 10;
	while (SERCOM3->SPI.SYNCBUSY.bit.CTRLB);
	SERCOM3->SPI.CTRLA.reg = ctrla = SERCOM_SPI_CTRLA_DIPO(3) | SERCOM_SPI_CTRLA_DOPO(0) | SERCOM_SPI_CTRLA_MODE_SPI_MASTER;
	SERCOM3->SPI.CTRLA.reg = ctrla | SERCOM_SPI_CTRLA_ENABLE;
	while (SERCOM3->SPI.SYNCBUSY.bit.ENABLE);
	
	return true;
}

static uint8_t sdSpiByte(uint_fast8_t val)
{
	while (!SERCOM3->SPI.INTFLAG.bit.DRE);
	SERCOM3->SPI.DATA.reg = val;
	while (!SERCOM3->SPI.INTFLAG.bit.RXC);
	return SERCOM3->SPI.DATA.reg;
}

static void sdChipSelect(void)
{
	PORT->Group[0].OUTCLR.reg = PORT_PA28;
}

static void sdChipDeselect(void)
{
	while (!SERCOM3->SPI.INTFLAG.bit.TXC);
	PORT->Group[0].OUTSET.reg = PORT_PA28;
}

static uint_fast8_t sdCrc7(uint8_t* chr, uint_fast8_t cnt, uint_fast8_t crc)
{
	uint_fast8_t i, a, data;

	for (a = 0; a < cnt; a++) {
		
		data = chr[a];
		
		for (i = 0; i < 8; i++, data <<= 1) {
			
			crc <<= 1;

			if ((data & 0x80) ^ (crc & 0x80))
				crc ^= 0x09;
		}
	}
	
	return crc & 0x7F;
}

static void sdPrvSendCmd(uint8_t cmd, uint32_t param, bool crc)
{
	sdSpiByte(0x40 | cmd);
	sdSpiByte(param >> 24);
	sdSpiByte(param >> 16);
	sdSpiByte(param >> 8);
	sdSpiByte(param);
	
	if (crc) {
		uint8_t send[] = {0x40 | cmd, param >> 24, param >> 16, param >> 8, param};
	
		sdSpiByte(sdCrc7(send, 5, 0) * 2 + 1);
	}
}

static uint_fast8_t sdPrvSimpleCommand(uint_fast8_t cmd, uint32_t param, bool crc, bool end_cs)	//do a command, return R1 reply
{
	uint_fast8_t ret, i = 0;
	
	sdChipSelect();
	
	sdPrvSendCmd(cmd, param, crc);
	
	do{		//our max wait time is 128 byte clocks (1024 clock ticks)
		
		ret = sdSpiByte(0xFF);
		
	} while (i++ < 128 && (ret == 0xFF));
	
	if (end_cs)
		sdChipDeselect();
	
	if (VERBOSE)
		pr("cmd %2xh (0x%08x) -> 0x%02x\n", cmd, param, ret);
	
	return ret;
}

static uint_fast8_t sdPrvACMD(uint_fast8_t cmd, uint32_t param, bool crc, bool end_cs)
{
	uint_fast8_t ret;
	
	ret = sdPrvSimpleCommand(55, 0, crc, true);
	if (ret & (FLAG_TIMEOUT | FLAG_ILLEGAL_CMD))
		return ret;
	
	return sdPrvSimpleCommand(cmd, param, crc, end_cs);
}

static bool sdPrvCardInit(bool isSdCard, bool signalHighCapacityHost)
{
	uint32_t param, time = 0, resp;
	bool first = true;
	
	param = (signalHighCapacityHost ? 0x40000000 : 0x00000000);
	
	while (time++ < 1024) {
	
		//busy bit is top bit of OCR, which is top bit of resp[1]
		//busy bit at 1 means init complete
		//see pages 85 and 26
		
		resp = isSdCard ? sdPrvACMD(41, param, true, true) : sdPrvSimpleCommand(1, param, true, true);
		
		if (resp & FLAG_TIMEOUT)
			break;
		
		if (!first && !(resp & FLAG_IN_IDLE_MODE))
			return true;
		first = false;
	}
	
	return false;
}

static uint_fast8_t sdPrvReadData(uint8_t* data, uint_fast16_t sz)
{
	uint_fast16_t tries = 2000;
	uint_fast8_t ret;
	
	do{
		ret = sdSpiByte(0xFF);
		if ((ret & 0xF0) == 0x00)
			return ret;	//fail
		if (ret == 0xFE)
			break;
		tries--;
	} while(tries);
	
	if (!tries)
		return 0xFF;
	
	*data = ret;
	ret = 0;
	
	while (sz--)
		*data++ = sdSpiByte(0xFF);
	
	//read crc
	sdSpiByte(0xFF);
	sdSpiByte(0xFF);
	
	return ret;
}

static uint32_t sdPrvGetBits(const uint8_t* data, uint32_t numBytesInArray, int32_t startBit, uint32_t len)	//for CID and CSD data..
{
	uint32_t ret = 0;
	int32_t i;
	
	for (i = startBit + len - 1; i >= startBit; i--)
		ret = (ret << 1) + ((data[numBytesInArray - 1 - i / 8] >> (i % 8)) & 1);
	
	return ret;
}

bool sdCardInit(void)
{
	uint8_t respBuf[16];
	uint_fast8_t i, v;
	uint32_t numSec;
	uint64_t t;
	
	if (!sdHwInit())
		return false;
	
	for (i = 0; i < 16; i++)	//give card time to init with CS deasserted
		sdSpiByte(0xff);
	
	v = sdPrvSimpleCommand(0, 0, true, true);
	if (v & FLAG_TIMEOUT)
		return false;
	if (!(v & FLAG_IN_IDLE_MODE))
		return false;
	
	v = sdPrvSimpleCommand(8, 0x000001AAUL, true, false);	//try CMD8 to init SDHC cards
	if (v & FLAG_TIMEOUT) {
		sdChipDeselect();
		return false;
	}
	mSD.HC = !(v & FLAG_ILLEGAL_CMD);
	if (mSD.HC) {
		sdSpiByte(0xFF);
		sdSpiByte(0xFF);
		sdSpiByte(0xFF);
		sdSpiByte(0xFF);
	}
	sdChipDeselect();
	
	v = sdPrvSimpleCommand(55, 0, true, true);			//see if this is SD or MMC
	if (v & FLAG_TIMEOUT)
		return false;
	mSD.SD = !(v & FLAG_ILLEGAL_CMD);
	
	if (mSD.SD) {
	
		if (!sdPrvCardInit(true, true) && !sdPrvCardInit(true, false))
			return false;
	}
	else{
	
		if (!sdPrvCardInit(false, true) && !sdPrvCardInit(false, false))
			return false;
	}
	
	v = sdPrvSimpleCommand(59, 0, true, true);		//crc off
	if (v & FLAG_TIMEOUT)
		return false;
		
	if (VERBOSE) {
	
		v = sdPrvSimpleCommand(9, 0, false, false);		//read CSD
		if (v & FLAG_TIMEOUT)
			return false;
		
		v = sdPrvReadData(respBuf, 16);
		sdChipDeselect();
		if (v)
			return false;
	
		pr("CSD:");
		for (i = 0; i < 16; i++)
			pr(" %02x", respBuf[i]);
		pr("\n\n");
		
		pr("CSD ver %xh\n", sdPrvGetBits(respBuf, 16, 126, 2) + 1);
	
		//calc capacity
		if (mSD.SD && sdPrvGetBits(respBuf, 16, 126, 2)) {	//v2 SD csd
			
			numSec = 1024 * (1 + sdPrvGetBits(respBuf, 16, 48, 22));
			
			pr("c_size: 0x%08x\n", sdPrvGetBits(respBuf, 16, 48, 22));
		}
		else {
			
			uint32_t cSize = sdPrvGetBits(respBuf, 16, 62, 12);
			uint32_t cSizeMult = sdPrvGetBits(respBuf, 16, 47, 3);
			uint32_t readBlLen = sdPrvGetBits(respBuf, 16, 80, 4);
			uint32_t shiftBy = readBlLen + cSizeMult + 2;
			
			if (!mSD.SD && sdPrvGetBits(respBuf, 16, 126, 2) >= 2 && sdPrvGetBits(respBuf, 16, 122, 4) >= 4) {
				
				pr("MMC+ over 4GB not supported due to lack of testing\n");
				
				return false;
			}
			
			if (shiftBy < 9) {	//fatal
				pr("card shift size invalid\n");
				return false;
			}
			shiftBy -= 9;
			numSec = (cSize + 1) << shiftBy;
		}
		
		pr("CARD capacity: %xh 512-byte blocks\n", numSec);

		if (mSD.SD) {
			
			v = sdPrvACMD(51, 0, false, false);
			if (v & FLAG_TIMEOUT)
				return false;
			
			v = sdPrvReadData(respBuf, 8);
			sdChipDeselect();
			if (v)
				return false;
				
			pr("SCR:");
			for (i = 0; i < 8; i++)
				pr(" %02x", respBuf[i]);
			pr("\n\n");
		}
	}
	
	//speed the clock up
	sdHwPrvSetBrg(0);
	
	//set block size
	v = sdPrvSimpleCommand(16, 512, true, true);		//crc off
	if (v & FLAG_TIMEOUT)
		return false;
	
	if (VERBOSE) {
		v = sdPrvSimpleCommand(10, 0, false, false);		//read CID
		if (v & FLAG_TIMEOUT)
			return false;
		
		v = sdPrvReadData(respBuf, 16);
		sdChipDeselect();
		if (v)
			return false;
	
		pr("CID:");
		for (i = 0; i < 16; i++)
			pr(" %02x", respBuf[i]);
		pr("\n\n");
		
		pr("serial number: %xh\n", sdPrvGetBits(respBuf, 16, 24, 32));
		pr("manuf: 0x%02x\n", sdPrvGetBits(respBuf, 16, 120, 8));
		pr("oem: 0x%04x\n", sdPrvGetBits(respBuf, 16, 104, 16));
		pr("prod: %c%c%c%c%c\n",
			sdPrvGetBits(respBuf, 16, 96, 8),
			sdPrvGetBits(respBuf, 16, 88, 8),
			sdPrvGetBits(respBuf, 16, 80, 8),
			sdPrvGetBits(respBuf, 16, 72, 8),
			sdPrvGetBits(respBuf, 16, 64, 8));
	}
	
	return true;
}

bool sdSecRead(uint32_t sec, uint8_t *dst)
{
	uint_fast8_t v;
	
	v = sdPrvSimpleCommand(17, mSD.HC ? sec : sec << 9, false, false);
	if(v & FLAG_TIMEOUT)
		return false;
	
	v = sdPrvReadData(dst, 512);
	sdChipDeselect();
	
	if (v)
		return false;
	
	return true;
}
	
	
	