
/*
	(c) 2021 Dmitry Grinberg   https://dmitry.gr
	Non-commercial use only OR licensing@dmitry.gr
*/

#include "samda1e16b.h"
#include "timebase.h"
#include <stdint.h>
#include <string.h>
#include <stddef.h>
#include "printf.h"
#include "atsamd.h"
#include "spiRam.h"

#define SET_MUX(_idx, _val)		do { if ((_idx) & 1) PORT_IOBUS->Group[0].PMUX[(_idx) / 2].bit.PMUXO = (_val); else PORT_IOBUS->Group[0].PMUX[(_idx) / 2].bit.PMUXE = (_val); } while(0)
	

#define QSPI_1_PINS		(PORT_PA00 | PORT_PA01 | PORT_PA02 | PORT_PA03)
#define QSPI_2_PINS		(PORT_PA04 | PORT_PA05 | PORT_PA06 | PORT_PA07)
#define QSPI_3_PINS		(PORT_PA08 | PORT_PA09 | PORT_PA10 | PORT_PA11)
#define QSPI_4_PINS		(PORT_PA14 | PORT_PA15 | PORT_PA16 | PORT_PA17)
#define QSPI_ALL_PINS	(QSPI_1_PINS | QSPI_2_PINS | QSPI_3_PINS | QSPI_4_PINS)
#define QSPI_ALL_MOSI	(PORT_PA00 | PORT_PA04 | PORT_PA08 | PORT_PA14)
#define MAX_NUM_CHIPS	(4)


#define SIZEOF_ID		(8)



#define MAX_CHIP_SIZE		(16)	//due to addressing
#define MIN_CHIP_SIZE		(1)


//no matter how i play it, 64-byte lines make things slower, not faster
#define CACHE_LINE_SIZE_ORDER	5		//must be at least the size of icache line, or else...
#define CACHE_NUM_WAYS			2		//number of lines a given PA can be in
#define CACHE_NUM_SETS			32		//number of buckets of PAs


#define CACHE_LINE_SIZE			(1 << CACHE_LINE_SIZE_ORDER)

#if OPTIMAL_RAM_WR_SZ > CACHE_LINE_SIZE || OPTIMAL_RAM_RD_SZ > CACHE_LINE_SIZE
	#error "you're in for a bad time"
#endif

#if CACHE_LINE_SIZE != 32
	#error "this is all tuned for 32 byte cache lines"
#endif

typedef void (*MemRead32Bfunc)(uint32_t addr, void *dstP);
typedef void (*MemWrite32Bfunc)(uint32_t addr, const void *srcP);

static uint8_t mTotalRam;
static MemRead32Bfunc mRead32B;
static MemWrite32Bfunc mWrite32B;



//cache code

#pragma GCC optimize ("Ofast")


struct CacheLine {
	uint32_t	dirty	:  1;
	uint32_t	addr	: 31;
	union {
		uint32_t dataW[CACHE_LINE_SIZE / sizeof(uint32_t)];
		uint16_t dataH[CACHE_LINE_SIZE / sizeof(uint16_t)];
		uint8_t dataB[CACHE_LINE_SIZE / sizeof(uint8_t)];
	};
};

struct CacheSet {
	struct CacheLine line[CACHE_NUM_WAYS];
};

struct Cache {
	struct CacheSet set[CACHE_NUM_SETS];
};

static struct Cache mCache;

static void spiRamCacheInit(void)
{
	uint_fast16_t way, set;
	
	for (set = 0; set < CACHE_NUM_SETS; set++) {
		for (way = 0; way < CACHE_NUM_WAYS; way++) {
			
			mCache.set[set].line[way].addr = 0x7fffffff;	//definitely invalid
		}
	}
}

static uint_fast16_t spiRamCachePrvHash(uint32_t addr)
{
	#if (CACHE_NUM_SETS & (CACHE_NUM_SETS - 1))
	
		#if CACHE_NUM_SETS == 3
			#define RECIP 	0xaaab
			#define SHIFT	17
		#elif CACHE_NUM_SETS == 5
			#define RECIP 	0xcccd
			#define SHIFT	18
		#elif CACHE_NUM_SETS == 6
			#define RECIP 	0xaaab
			#define SHIFT	18
		#elif CACHE_NUM_SETS == 9
			#define RECIP 	0xe38f
			#define SHIFT	19
		#elif CACHE_NUM_SETS == 10
			#define RECIP 	0xcccd
			#define SHIFT	19
		#elif CACHE_NUM_SETS == 11
			#define RECIP 	0xba2f
			#define SHIFT	19
		#elif CACHE_NUM_SETS == 12
			#define RECIP 	0xaaab
			#define SHIFT	19
		#elif CACHE_NUM_SETS == 13
			#define RECIP 	0x9d8a
			#define SHIFT	19
		#elif CACHE_NUM_SETS == 15
			#define RECIP 	0x8889
			#define SHIFT	19
		#elif CACHE_NUM_SETS == 18
			#define RECIP 	0xe38f
			#define SHIFT	20
		#elif CACHE_NUM_SETS == 20
			#define RECIP 	0xcccd
			#define SHIFT	20
		#elif CACHE_NUM_SETS == 22
			#define RECIP 	0xba2f
			#define SHIFT	20
		#elif CACHE_NUM_SETS == 24
			#define RECIP 	0xaaab
			#define SHIFT	20
		#elif CACHE_NUM_SETS == 26
			#define RECIP 	0x9d8a
			#define SHIFT	20
		#elif CACHE_NUM_SETS == 30
			#define RECIP 	0x8889
			#define SHIFT	20
		#else
			#error "we lack a reciprocal for this value - it will be slow. refusing!"
		#endif
		
		uint32_t div;
		
		addr /= CACHE_LINE_SIZE;
		addr = (uint16_t)addr;
		
		div = addr * RECIP;
		div >>= SHIFT;
		addr -= div * CACHE_NUM_SETS;
		
	#else
	
		addr /= CACHE_LINE_SIZE;
		addr %= CACHE_NUM_SETS;
	
	#endif
	
	return addr;
}

static uint_fast16_t spiRamCachePrvPickVictim(struct Cache *cache)
{
	return SysTick->VAL % CACHE_NUM_WAYS;
}

static struct CacheLine* spiRamCachePrvFillLine(struct Cache *cache, struct CacheSet *set, uint32_t addr, bool loadFromRam)
{
	uint_fast16_t idx = spiRamCachePrvPickVictim(cache);
	struct CacheLine *line = &set->line[idx];
	
//	pr("picked victim way %u currently holding addr 0x%08x, %s\n", idx, line->addr * CACHE_LINE_SIZE, line->dirty ? "DIRTY" : "CLEAN");
	if (line->dirty) {	//clean line
		

//		pr(" flushing -> %08x\n", line->addr * CACHE_LINE_SIZE);
		mWrite32B(line->addr * CACHE_LINE_SIZE, line->dataW);
	}
	
	if (loadFromRam) {
//		uint32_t *dst = line->dataW;

		mRead32B(addr / CACHE_LINE_SIZE * CACHE_LINE_SIZE, line->dataW);
		
//		pr(" filled %08x -> %08x %08x %08x %08x %08x %08x %08x %08x\n", addr, 
//				dst[0], dst[1], dst[2], dst[3], dst[4], dst[5], dst[6], dst[7]);
	}
	line->dirty = 0;
	line->addr = addr / CACHE_LINE_SIZE;
	
	return line;
}

#if CACHE_NUM_WAYS == 2
//this code hardwired for 2 ways, can be adjusted

	#if CACHE_NUM_SETS == 32
		#define HASH_N_STUFF							\
			"   lsls  r4, #27					\n\t"	\
			"   lsrs  r4, #27					\n\t"
	#elif CACHE_NUM_SETS == 30
		#define HASH_N_STUFF							\
			"	uxth  r4, r4					\n\t"	\
			"	ldr   r5, =0x8889				\n\t"	\
			"	muls  r5, r4					\n\t"	\
			"	lsrs  r5, #20					\n\t"	\
			"	movs  r3, #30					\n\t"	\
			"	muls  r3, r5					\n\t"	\
			"	subs  r4, r3					\n\t"
	#elif CACHE_NUM_SETS == 26
		#define HASH_N_STUFF							\
			"	uxth  r4, r4					\n\t"	\
			"	ldr   r5, =0x9d8a				\n\t"	\
			"	muls  r5, r4					\n\t"	\
			"	lsrs  r5, #20					\n\t"	\
			"	movs  r3, #26					\n\t"	\
			"	muls  r3, r5					\n\t"	\
			"	subs  r4, r3					\n\t"
	#elif CACHE_NUM_SETS == 24
		#define HASH_N_STUFF							\
			"	uxth  r4, r4					\n\t"	\
			"	ldr   r5, =0xaaab				\n\t"	\
			"	muls  r5, r4					\n\t"	\
			"	lsrs  r5, #20					\n\t"	\
			"	movs  r3, #24					\n\t"	\
			"	muls  r3, r5					\n\t"	\
			"	subs  r4, r3					\n\t"
	#elif CACHE_NUM_SETS == 22
		#define HASH_N_STUFF							\
			"	uxth  r4, r4					\n\t"	\
			"	ldr   r5, =0xba2f				\n\t"	\
			"	muls  r5, r4					\n\t"	\
			"	lsrs  r5, #20					\n\t"	\
			"	movs  r3, #22					\n\t"	\
			"	muls  r3, r5					\n\t"	\
			"	subs  r4, r3					\n\t"
	#elif CACHE_NUM_SETS == 20
		#define HASH_N_STUFF							\
			"	uxth  r4, r4					\n\t"	\
			"	ldr   r5, =0xcccd				\n\t"	\
			"	muls  r5, r4					\n\t"	\
			"	lsrs  r5, #20					\n\t"	\
			"	movs  r3, #20					\n\t"	\
			"	muls  r3, r5					\n\t"	\
			"	subs  r4, r3					\n\t"
	#elif CACHE_NUM_SETS == 18
		#define HASH_N_STUFF							\
			"	uxth  r4, r4					\n\t"	\
			"	ldr   r5, =0xe38f				\n\t"	\
			"	muls  r5, r4					\n\t"	\
			"	lsrs  r5, #20					\n\t"	\
			"	movs  r3, #18					\n\t"	\
			"	muls  r3, r5					\n\t"	\
			"	subs  r4, r3					\n\t"
	#elif CACHE_NUM_SETS == 16
		#define HASH_N_STUFF							\
			"   lsls  r4, #28					\n\t"	\
			"   lsrs  r4, #28					\n\t"
	#elif CACHE_NUM_SETS == 15
		#define HASH_N_STUFF							\
			"	uxth  r4, r4					\n\t"	\
			"	ldr   r5, =0x8889				\n\t"	\
			"	muls  r5, r4					\n\t"	\
			"	lsrs  r5, #19					\n\t"	\
			"	movs  r3, #15					\n\t"	\
			"	muls  r3, r5					\n\t"	\
			"	subs  r4, r3					\n\t"
	#elif CACHE_NUM_SETS == 13
		#define HASH_N_STUFF							\
			"	uxth  r4, r4					\n\t"	\
			"	ldr   r5, =0x9d8a				\n\t"	\
			"	muls  r5, r4					\n\t"	\
			"	lsrs  r5, #19					\n\t"	\
			"	movs  r3, #13					\n\t"	\
			"	muls  r3, r5					\n\t"	\
			"	subs  r4, r3					\n\t"
	#elif CACHE_NUM_SETS == 12
		#define HASH_N_STUFF							\
			"	uxth  r4, r4					\n\t"	\
			"	ldr   r5, =0xaaab				\n\t"	\
			"	muls  r5, r4					\n\t"	\
			"	lsrs  r5, #19					\n\t"	\
			"	movs  r3, #12					\n\t"	\
			"	muls  r3, r5					\n\t"	\
			"	subs  r4, r3					\n\t"
	#elif CACHE_NUM_SETS == 8
		#define HASH_N_STUFF							\
			"   lsls  r4, #29					\n\t"	\
			"   lsrs  r4, #29					\n\t"
	#else
	
		#error "bad setting for fast path"
	#endif




	void __attribute__((naked, noinline)) spiRamRead(uint32_t addr, void *dataP, uint_fast16_t sz)
	{
		asm volatile(
			".syntax unified					\n\t"
			"	push  {r4, r5}					\n\t"
			"	lsrs  r4, r0, %2				\n\t"	//r4 = addr / CACHE_LINE_SIZE
			
			HASH_N_STUFF
			
			"	ldr   r3, =%0					\n\t"
			"	ldr   r5, =%3 * 2 + 4 * 2		\n\t"	//size of each set
			"	muls  r4, r5					\n\t"
			"	adds  r4, r3					\n\t"	//points to proper set
			"	ldmia r4!, {r5}					\n\t"	//get info[0], increment past it
			"	lsrs  r3, r0, %2				\n\t"	//addr to compare to
			"	lsrs  r5, #1					\n\t"	//hide dirty bit
			"	cmp   r3, r5					\n\t"
			"	beq   1f						\n\t"
			"	ldr   r5, [r4, %3]				\n\t"	//get info [1]
			"	lsrs  r5, #1					\n\t"	//hide dirty bit
			"	cmp   r3, r5					\n\t"
			"	bne   2f						\n\t"	//not found
			//found in set 2
			"	adds  r4, %3 + 4				\n\t"
			"1:									\n\t"	//common path for "found"
			"	lsls  r0, %4					\n\t"
			"	lsrs  r0, %4					\n\t"
			"	cmp   r2, #4					\n\t"
			"	bne   5f						\n\t"
			"	ldr   r5, [r4, r0]				\n\t"
			"	str   r5, [r1]					\n\t"
			"	pop   {r4, r5}					\n\t"
			"	bx    lr						\n\t"
			"5:									\n\t"
			"	cmp   r2, #1					\n\t"
			"	bne   5f						\n\t"
			"	ldrb  r5, [r4, r0]				\n\t"
			"	strb  r5, [r1]					\n\t"
			"	pop   {r4, r5}					\n\t"
			"	bx    lr						\n\t"
			"5:									\n\t"
			"	cmp   r2, #2					\n\t"
			"	bne   5f						\n\t"
			"	ldrh  r5, [r4, r0]				\n\t"
			"	strh  r5, [r1]					\n\t"
			"	pop   {r4, r5}					\n\t"
			"	bx    lr						\n\t"
			"5:									\n\t"
			"	cmp   r2, #32					\n\t"
			"	bne   5f						\n\t"
			"	adds  r4, r0					\n\t"
			"	ldmia r4!, {r0, r2, r3, r5}		\n\t"
			"	stmia r1!, {r0, r2, r3, r5}		\n\t"
			"	ldmia r4!, {r0, r2, r3, r5}		\n\t"
			"	stmia r1!, {r0, r2, r3, r5}		\n\t"
			"	pop   {r4, r5}					\n\t"
			"	bx    lr						\n\t"
			"5:									\n\t"
			"	cmp   r2, #64					\n\t"
			"	bne   5f						\n\t"
			"	adds  r4, r0					\n\t"
			"	ldmia r4!, {r0, r2, r3, r5}		\n\t"
			"	stmia r1!, {r0, r2, r3, r5}		\n\t"
			"	ldmia r4!, {r0, r2, r3, r5}		\n\t"
			"	stmia r1!, {r0, r2, r3, r5}		\n\t"
			"	ldmia r4!, {r0, r2, r3, r5}		\n\t"
			"	stmia r1!, {r0, r2, r3, r5}		\n\t"
			"	ldmia r4!, {r0, r2, r3, r5}		\n\t"
			"	stmia r1!, {r0, r2, r3, r5}		\n\t"
			"	pop   {r4, r5}					\n\t"
			"	bx    lr						\n\t"
	
			"5: 								\n\t"		//not supported - use slow path
			"	pop   {r4, r5}					\n\t"
			"	b     3f						\n\t"
			
			"2:									\n\t"	//not found
			"	pop   {r4, r5}					\n\t"
			"	cmp   r2, %3					\n\t"
			"	bne   3f						\n\t"	//slow path
			//direct read path					\n\t"
			"	ldr   r3, =%1					\n\t"
			"	ldr   r3, [r3]					\n\t"
			"	bx    r3						\n\t"
			"3:									\n\t"	//slow
			"	ldr   r3, =spiRamRead_slowpath	\n\t"
			"	bx    r3						\n\t"
			:
			:"i"(mCache.set), "i"(&mRead32B), "i"(CACHE_LINE_SIZE_ORDER), "i"(CACHE_LINE_SIZE), "i"(32 - CACHE_LINE_SIZE_ORDER)
		);
	}
	
	
	void __attribute__((used)) spiRamRead_slowpath(uint32_t addr, void *dataP, uint_fast16_t sz)
	{
		struct Cache *cache = &mCache;
		struct CacheSet *set = &cache->set[spiRamCachePrvHash(addr)];
		uint32_t *dptr, *dst = (uint32_t*)dataP, dummy1, dummy2;
		struct CacheLine *line; 
	
	//	pr("slow read %u @ 0x%08x -> set %u\n", sz, addr, set - &cache->set[0]);
		
	//	pr("fill\n", sz, addr);
		line = spiRamCachePrvFillLine(cache, set, addr, true);
	
		
		switch (sz) {
			case 1:
				*(uint8_t*)dataP = line->dataB[(addr % CACHE_LINE_SIZE) / sizeof(uint8_t)];
	//			pr("read %u @ 0x%08x idx %u -> %02x\n", sz, addr, line - &set->line[0], *(uint8_t*)dataP);
				break;
			
			case 2:
				*(uint16_t*)dataP = line->dataH[(addr % CACHE_LINE_SIZE) / sizeof(uint16_t)];
	//			pr("read %u @ 0x%08x idx %u -> %04x\n", sz, addr, line - &set->line[0], *(uint16_t*)dataP);
				break;
			
			case 4:
				*(uint32_t*)dataP = line->dataW[(addr % CACHE_LINE_SIZE) / sizeof(uint32_t)];
	//			pr("read %u @ 0x%08x idx %u -> %08x\n", sz, addr, line - &set->line[0], *(uint32_t*)dataP);
				break;
			
			case 8:
				dptr = &line->dataW[(addr % CACHE_LINE_SIZE) / sizeof(uint32_t)];
				asm volatile(
					"	ldmia %0!, {r0, r1}	\n"
					"	stmia %1!, {r0, r1}	\n"
					:"=l"(dummy1), "=l"(dummy2)
					:"0"(dptr), "1"(dst)
					:"memory", "r0", "r1"
				);
	//			pr("read %u @ 0x%08x idx %u -> %08x %08x\n", sz, addr, line - &set->line[0], 
	//				dst[0], dst[1]);
				break;
			
			case 16:
				dptr = &line->dataW[(addr % CACHE_LINE_SIZE) / sizeof(uint32_t)];
				asm volatile(
					"	ldmia %0!, {r0, r1, r2, r3}	\n"
					"	stmia %1!, {r0, r1, r2, r3}	\n"
					:"=l"(dummy1), "=l"(dummy2)
					:"0"(dptr), "1"(dst)
					:"memory", "r0", "r1", "r2", "r3"
				);
	//			pr("read %u @ 0x%08x idx %u -> %08x %08x %08x %08x\n", sz, addr, line - &set->line[0], 
	//				dst[0], dst[1], dst[2], dst[3]);
				break;
	
	#if CACHE_LINE_SIZE >= 32
			case 32:
				dptr = &line->dataW[(addr % CACHE_LINE_SIZE) / sizeof(uint32_t)];
				asm volatile(
					"	ldmia %0!, {r0, r1, r2, r3}	\n"
					"	stmia %1!, {r0, r1, r2, r3}	\n"
					"	ldmia %0!, {r0, r1, r2, r3}	\n"
					"	stmia %1!, {r0, r1, r2, r3}	\n"
					:"=l"(dummy1), "=l"(dummy2)
					:"0"(dptr), "1"(dst)
					:"memory", "r0", "r1", "r2", "r3"
				);
	//			pr("read %u @ 0x%08x idx %u -> %08x %08x %08x %08x %08x %08x %08x %08x\n", sz, addr, line - &set->line[0], 
	//				dst[0], dst[1], dst[2], dst[3], dst[4], dst[5], dst[6], dst[7]);
				break;
	#endif
	#if CACHE_LINE_SIZE >= 64
			case 64:
				dptr = &line->dataW[(addr % CACHE_LINE_SIZE) / sizeof(uint32_t)];
				asm volatile(
					"	ldmia %0!, {r0, r1, r2, r3}	\n"
					"	stmia %1!, {r0, r1, r2, r3}	\n"
					"	ldmia %0!, {r0, r1, r2, r3}	\n"
					"	stmia %1!, {r0, r1, r2, r3}	\n"
					"	ldmia %0!, {r0, r1, r2, r3}	\n"
					"	stmia %1!, {r0, r1, r2, r3}	\n"
					"	ldmia %0!, {r0, r1, r2, r3}	\n"
					"	stmia %1!, {r0, r1, r2, r3}	\n"
					:"=l"(dummy1), "=l"(dummy2)
					:"0"(dptr), "1"(dst)
					:"memory", "r0", "r1", "r2", "r3"
				);
	//			pr("read %u @ 0x%08x idx %u -> %08x %08x %08x %08x %08x %08x %08x %08x %08x %08x %08x %08x %08x %08x %08x %08x\n", sz, addr, line - &set->line[0], 
	//				dst[0], dst[1], dst[2], dst[3], dst[4], dst[5], dst[6], dst[7],
	//				dst[8], dst[9], dst[10], dst[11], dst[12], dst[13], dst[14], dst[15]);
				break;
	#endif
	
			default:
				pr("unknown size %u\n", sz);
				while(1);
		}
	}

	void __attribute__((naked, noinline)) spiRamWrite(uint32_t addr, const void *dataP, uint_fast16_t sz)
	{
		asm volatile(
			".syntax unified					\n\t"
			"	push  {r4, r5}					\n\t"
			"	lsrs  r4, r0, %2				\n\t"	//r4 = addr / CACHE_LINE_SIZE
			
			HASH_N_STUFF
			
			"	ldr   r3, =%0					\n\t"
			"	ldr   r5, =%3 * 2 + 4 * 2		\n\t"	//size of each set
			"	muls  r4, r5					\n\t"
			"	adds  r4, r3					\n\t"	//points to proper set
			"	ldmia r4!, {r5}					\n\t"	//get info[0], increment past it
			"	lsrs  r3, r0, %2				\n\t"	//addr to compare to
			"	lsrs  r5, #1					\n\t"	//hide dirty bit
			"	cmp   r3, r5					\n\t"
			"	beq   1f						\n\t"
			"	ldr   r5, [r4, %3]				\n\t"	//get info [1]
			"	lsrs  r5, #1					\n\t"	//hide dirty bit
			"	cmp   r3, r5					\n\t"
			"	bne   2f						\n\t"	//not found
			//found in set 2
			"	adds  r4, %3 + 4				\n\t"
			"1:									\n\t"	//common path for "found"
			"	lsls  r0, %4					\n\t"
			"	lsrs  r0, %4					\n\t"
			"	cmp   r2, #4					\n\t"
			"	bne   5f						\n\t"
			"	ldr   r5, [r1]					\n\t"
			"	str   r5, [r4, r0]				\n\t"

			"4:									\n\t"		//mark as dirty
			"	subs  r4, #4					\n\t"
			"	ldr   r0, [r4]					\n\t"
			"	movs  r1, #1					\n\t"
			"	orrs  r0, r1					\n\t"
			"	str   r0, [r4]					\n\t"
			"	pop   {r4, r5}					\n\t"
			"	bx    lr						\n\t"

			"5:									\n\t"
			"	cmp   r2, #1					\n\t"
			"	bne   5f						\n\t"
			"	ldrb  r5, [r1]					\n\t"
			"	strb  r5, [r4, r0]				\n\t"
			"	b     4b						\n\t"
			"5:									\n\t"
			"	cmp   r2, #2					\n\t"
			"	bne   5f						\n\t"
			"	ldrh  r5, [r1]					\n\t"
			"	strh  r5, [r4, r0]				\n\t"
			"	b     4b						\n\t"
			"5:									\n\t"
			"	cmp   r2, #32					\n\t"
			"	bne   5f						\n\t"
			"	mov   r12, r4					\n\t"
			"	adds  r4, r0					\n\t"
			"	ldmia r1!, {r0, r2, r3, r5}		\n\t"
			"	stmia r4!, {r0, r2, r3, r5}		\n\t"
			"	ldmia r1!, {r0, r2, r3, r5}		\n\t"
			"	stmia r4!, {r0, r2, r3, r5}		\n\t"
			"	mov   r4, r12					\n\t"
			"	b     4b						\n\t"
			"5:									\n\t"
			"	cmp   r2, #64					\n\t"
			"	bne   5f						\n\t"
			"	mov   r12, r4					\n\t"
			"	adds  r4, r0					\n\t"
			"	ldmia r1!, {r0, r2, r3, r5}		\n\t"
			"	stmia r4!, {r0, r2, r3, r5}		\n\t"
			"	ldmia r1!, {r0, r2, r3, r5}		\n\t"
			"	stmia r4!, {r0, r2, r3, r5}		\n\t"
			"	ldmia r1!, {r0, r2, r3, r5}		\n\t"
			"	stmia r4!, {r0, r2, r3, r5}		\n\t"
			"	ldmia r1!, {r0, r2, r3, r5}		\n\t"
			"	stmia r4!, {r0, r2, r3, r5}		\n\t"
			"	mov   r4, r12					\n\t"
			"	b     4b						\n\t"
	
			"5: 								\n\t"		//not supported - use slow path
			"	pop   {r4, r5}					\n\t"
			"	b     3f						\n\t"
		
			"2:									\n\t"	//not found
			"	pop   {r4, r5}					\n\t"
			"	cmp   r2, %3					\n\t"
			"	bne   3f						\n\t"	//slow path
			//direct read path					\n\t"
			"	ldr   r3, =%1					\n\t"
			"	ldr   r3, [r3]					\n\t"
			"	bx    r3						\n\t"
			"3:									\n\t"	//slow
			"	ldr   r3, =spiRamWrite_slowpath	\n\t"
			"	bx    r3						\n\t"
			:
			:"i"(mCache.set), "i"(&mWrite32B), "i"(CACHE_LINE_SIZE_ORDER), "i"(CACHE_LINE_SIZE), "i"(32 - CACHE_LINE_SIZE_ORDER)
			:"memory", "cc"
		);
	}
	

	void __attribute__((used)) spiRamWrite_slowpath(uint32_t addr, const void *dataP, uint_fast16_t sz)
	{
		struct Cache *cache = &mCache;
		struct CacheSet *set = &cache->set[spiRamCachePrvHash(addr)];
		const uint32_t *src = (const uint32_t*)dataP;
		uint32_t dummy1, dummy2;
		struct CacheLine *line; 
		uint32_t *dptr;
		
		
	//	pr("slow write %u @ 0x%08x -> set %u\n", sz, addr, set - cache->set);
		
		//not found
	//	pr("fill\n", sz, addr);
		line = spiRamCachePrvFillLine(cache, set, addr, sz != CACHE_LINE_SIZE);
		
		switch (sz) {
			case 1:
	//			pr("write %u @ 0x%08x idx %u <- %02x\n", sz, addr, line - &set->line[0], *(uint8_t*)dataP);
				line->dataB[(addr % CACHE_LINE_SIZE) / sizeof(uint8_t)] = *(uint8_t*)dataP;
				break;
			
			case 2:
	//			pr("write %u @ 0x%08x idx %u <- %04x\n", sz, addr, line - &set->line[0], *(uint16_t*)dataP);
				line->dataH[(addr % CACHE_LINE_SIZE) / sizeof(uint16_t)] = *(uint16_t*)dataP;
				break;
			
			case 4:
	//			pr("write %u @ 0x%08x idx %u <- %08x\n", sz, addr, line - &set->line[0], *(uint32_t*)dataP);
				line->dataW[(addr % CACHE_LINE_SIZE) / sizeof(uint32_t)] = *(uint32_t*)dataP;
				break;
			
			case 8:
	//			pr("write %u @ 0x%08x idx %u <- %08x %08x\n", sz, addr, line - &set->line[0], 
	//				src[0], src[1]);
				dptr = &line->dataW[(addr % CACHE_LINE_SIZE) / sizeof(uint32_t)];
				asm volatile(
					"	ldmia %0!, {r0, r1}	\n"
					"	stmia %1!, {r0, r1}	\n"
					:"=l"(dummy1), "=l"(dummy2)
					:"0"(src), "1"(dptr)
					:"memory", "r0", "r1"
				);
				break;
			
			case 16:
	//			pr("write %u @ 0x%08x idx %u <- %08x %08x %08x %08x\n", sz, addr, line - &set->line[0], 
	//				src[0], src[1], src[2], src[3]);
				dptr = &line->dataW[(addr % CACHE_LINE_SIZE) / sizeof(uint32_t)];
				asm volatile(
					"	ldmia %0!, {r0, r1, r2, r3}	\n"
					"	stmia %1!, {r0, r1, r2, r3}	\n"
					:"=l"(dummy1), "=l"(dummy2)
					:"0"(src), "1"(dptr)
					:"memory", "r0", "r1", "r2", "r3"
				);
				break;
	#if CACHE_LINE_SIZE >= 32
			case 32:
	//			pr("write %u @ 0x%08x idx %u <- %08x %08x %08x %08x %08x %08x %08x %08x\n", sz, addr, line - &set->line[0], 
	//				src[0], src[1], src[2], src[3], src[4], src[5], src[6], src[7]);
				dptr = &line->dataW[(addr % CACHE_LINE_SIZE) / sizeof(uint32_t)];
				asm volatile(
					"	ldmia %0!, {r0, r1, r2, r3}	\n"
					"	stmia %1!, {r0, r1, r2, r3}	\n"
					"	ldmia %0!, {r0, r1, r2, r3}	\n"
					"	stmia %1!, {r0, r1, r2, r3}	\n"
					:"=l"(dummy1), "=l"(dummy2)
					:"0"(src), "1"(dptr)
					:"memory", "r0", "r1", "r2", "r3"
				);
				break;
	#endif
	#if CACHE_LINE_SIZE >= 64
			case 64:
	//			pr("write %u @ 0x%08x idx %u <- %08x %08x %08x %08x %08x %08x %08x %08x %08x %08x %08x %08x %08x %08x %08x %08x\n", sz, addr, line - &set->line[0], 
	//				src[0], src[1], src[2], src[3], src[4], src[5], src[6], src[7],
	//				src[8], src[9], src[10], src[11], src[12], src[13], src[14], src[15]);
				dptr = &line->dataW[(addr % CACHE_LINE_SIZE) / sizeof(uint32_t)];
				asm volatile(
					"	ldmia %0!, {r0, r1, r2, r3}	\n"
					"	stmia %1!, {r0, r1, r2, r3}	\n"
					"	ldmia %0!, {r0, r1, r2, r3}	\n"
					"	stmia %1!, {r0, r1, r2, r3}	\n"
					"	ldmia %0!, {r0, r1, r2, r3}	\n"
					"	stmia %1!, {r0, r1, r2, r3}	\n"
					:"=l"(dummy1), "=l"(dummy2)
					:"0"(src), "1"(dptr)
					:"memory", "r0", "r1", "r2", "r3"
				);
				break;
	#endif
			default:
				pr("unknown size %u\n", sz);
				while(1);
		}
		line->dirty = 1;
	}


#else
	
	#warning "using slow path";
	
	void spiRamRead(uint32_t addr, void *dataP, uint_fast16_t sz)	//assume not called with zero
	{
		struct Cache *cache = &mCache;
		struct CacheSet *set = &cache->set[spiRamCachePrvHash(addr)];
		uint32_t *dptr, *dst = (uint32_t*)dataP, dummy1, dummy2;
		struct CacheLine *line; 
		uint_fast16_t i;
	
	//	pr("read %u @ 0x%08x -> set %u\n", sz, addr, set - &cache->set[0]);
		
		for (i = 0, line = &set->line[0]; i < CACHE_NUM_WAYS; i++, line++) {
			
			if (line->addr == addr / CACHE_LINE_SIZE)
				goto found;
		}
		
		
		//not found
		
		if (sz == CACHE_LINE_SIZE) {
			
			//no point allocating it
			mRead32B(addr, dataP);
			return;
		}
			
	//	pr("fill\n", sz, addr);
		line = spiRamCachePrvFillLine(cache, set, addr, true);
	
	found:
		
		switch (sz) {
			case 1:
				*(uint8_t*)dataP = line->dataB[(addr % CACHE_LINE_SIZE) / sizeof(uint8_t)];
	//			pr("read %u @ 0x%08x idx %u -> %02x\n", sz, addr, line - &set->line[0], *(uint8_t*)dataP);
				break;
			
			case 2:
				*(uint16_t*)dataP = line->dataH[(addr % CACHE_LINE_SIZE) / sizeof(uint16_t)];
	//			pr("read %u @ 0x%08x idx %u -> %04x\n", sz, addr, line - &set->line[0], *(uint16_t*)dataP);
				break;
			
			case 4:
				*(uint32_t*)dataP = line->dataW[(addr % CACHE_LINE_SIZE) / sizeof(uint32_t)];
	//			pr("read %u @ 0x%08x idx %u -> %08x\n", sz, addr, line - &set->line[0], *(uint32_t*)dataP);
				break;
			
			case 8:
				dptr = &line->dataW[(addr % CACHE_LINE_SIZE) / sizeof(uint32_t)];
				asm volatile(
					"	ldmia %0!, {r0, r1}	\n"
					"	stmia %1!, {r0, r1}	\n"
					:"=l"(dummy1), "=l"(dummy2)
					:"0"(dptr), "1"(dst)
					:"memory", "r0", "r1"
				);
	//			pr("read %u @ 0x%08x idx %u -> %08x %08x\n", sz, addr, line - &set->line[0], 
	//				dst[0], dst[1]);
				break;
			
			case 16:
				dptr = &line->dataW[(addr % CACHE_LINE_SIZE) / sizeof(uint32_t)];
				asm volatile(
					"	ldmia %0!, {r0, r1, r2, r3}	\n"
					"	stmia %1!, {r0, r1, r2, r3}	\n"
					:"=l"(dummy1), "=l"(dummy2)
					:"0"(dptr), "1"(dst)
					:"memory", "r0", "r1", "r2", "r3"
				);
	//			pr("read %u @ 0x%08x idx %u -> %08x %08x %08x %08x\n", sz, addr, line - &set->line[0], 
	//				dst[0], dst[1], dst[2], dst[3]);
				break;
	
	#if CACHE_LINE_SIZE >= 32
			case 32:
				dptr = &line->dataW[(addr % CACHE_LINE_SIZE) / sizeof(uint32_t)];
				asm volatile(
					"	ldmia %0!, {r0, r1, r2, r3}	\n"
					"	stmia %1!, {r0, r1, r2, r3}	\n"
					"	ldmia %0!, {r0, r1, r2, r3}	\n"
					"	stmia %1!, {r0, r1, r2, r3}	\n"
					:"=l"(dummy1), "=l"(dummy2)
					:"0"(dptr), "1"(dst)
					:"memory", "r0", "r1", "r2", "r3"
				);
	//			pr("read %u @ 0x%08x idx %u -> %08x %08x %08x %08x %08x %08x %08x %08x\n", sz, addr, line - &set->line[0], 
	//				dst[0], dst[1], dst[2], dst[3], dst[4], dst[5], dst[6], dst[7]);
				break;
	#endif
	#if CACHE_LINE_SIZE >= 64
			case 64:
				dptr = &line->dataW[(addr % CACHE_LINE_SIZE) / sizeof(uint32_t)];
				asm volatile(
					"	ldmia %0!, {r0, r1, r2, r3}	\n"
					"	stmia %1!, {r0, r1, r2, r3}	\n"
					"	ldmia %0!, {r0, r1, r2, r3}	\n"
					"	stmia %1!, {r0, r1, r2, r3}	\n"
					"	ldmia %0!, {r0, r1, r2, r3}	\n"
					"	stmia %1!, {r0, r1, r2, r3}	\n"
					"	ldmia %0!, {r0, r1, r2, r3}	\n"
					"	stmia %1!, {r0, r1, r2, r3}	\n"
					:"=l"(dummy1), "=l"(dummy2)
					:"0"(dptr), "1"(dst)
					:"memory", "r0", "r1", "r2", "r3"
				);
	//			pr("read %u @ 0x%08x idx %u -> %08x %08x %08x %08x %08x %08x %08x %08x %08x %08x %08x %08x %08x %08x %08x %08x\n", sz, addr, line - &set->line[0], 
	//				dst[0], dst[1], dst[2], dst[3], dst[4], dst[5], dst[6], dst[7],
	//				dst[8], dst[9], dst[10], dst[11], dst[12], dst[13], dst[14], dst[15]);
				break;
	#endif
	
			default:
				pr("unknown size %u\n", sz);
				while(1);
		}
	}

	void spiRamWrite(uint32_t addr, const void *dataP, uint_fast16_t sz)
	{
		struct Cache *cache = &mCache;
		struct CacheSet *set = &cache->set[spiRamCachePrvHash(addr)];
		const uint32_t *src = (const uint32_t*)dataP;
		uint32_t dummy1, dummy2;
		struct CacheLine *line; 
		uint32_t *dptr;
		uint_fast16_t i;
		
		
	//	pr("write %u @ 0x%08x -> set %u\n", sz, addr, set - cache->set);
		
		for (i = 0, line = &set->line[0]; i < CACHE_NUM_WAYS; i++, line++) {
			
			if (line->addr == addr / CACHE_LINE_SIZE)
				goto found;
		}
		
		
		if (sz == CACHE_LINE_SIZE)	//missed writes of cache line size or more
			return mWrite32B(addr, dataP);
			
		//not found
	//	pr("fill\n", sz, addr);
		line = spiRamCachePrvFillLine(cache, set, addr, sz != CACHE_LINE_SIZE);
	
	found:
		
		switch (sz) {
			case 1:
	//			pr("write %u @ 0x%08x idx %u <- %02x\n", sz, addr, line - &set->line[0], *(uint8_t*)dataP);
				line->dataB[(addr % CACHE_LINE_SIZE) / sizeof(uint8_t)] = *(uint8_t*)dataP;
				break;
			
			case 2:
	//			pr("write %u @ 0x%08x idx %u <- %04x\n", sz, addr, line - &set->line[0], *(uint16_t*)dataP);
				line->dataH[(addr % CACHE_LINE_SIZE) / sizeof(uint16_t)] = *(uint16_t*)dataP;
				break;
			
			case 4:
	//			pr("write %u @ 0x%08x idx %u <- %08x\n", sz, addr, line - &set->line[0], *(uint32_t*)dataP);
				line->dataW[(addr % CACHE_LINE_SIZE) / sizeof(uint32_t)] = *(uint32_t*)dataP;
				break;
			
			case 8:
	//			pr("write %u @ 0x%08x idx %u <- %08x %08x\n", sz, addr, line - &set->line[0], 
	//				src[0], src[1]);
				dptr = &line->dataW[(addr % CACHE_LINE_SIZE) / sizeof(uint32_t)];
				asm volatile(
					"	ldmia %0!, {r0, r1}	\n"
					"	stmia %1!, {r0, r1}	\n"
					:"=l"(dummy1), "=l"(dummy2)
					:"0"(src), "1"(dptr)
					:"memory", "r0", "r1"
				);
				break;
			
			case 16:
	//			pr("write %u @ 0x%08x idx %u <- %08x %08x %08x %08x\n", sz, addr, line - &set->line[0], 
	//				src[0], src[1], src[2], src[3]);
				dptr = &line->dataW[(addr % CACHE_LINE_SIZE) / sizeof(uint32_t)];
				asm volatile(
					"	ldmia %0!, {r0, r1, r2, r3}	\n"
					"	stmia %1!, {r0, r1, r2, r3}	\n"
					:"=l"(dummy1), "=l"(dummy2)
					:"0"(src), "1"(dptr)
					:"memory", "r0", "r1", "r2", "r3"
				);
				break;
	#if CACHE_LINE_SIZE >= 32
			case 32:
	//			pr("write %u @ 0x%08x idx %u <- %08x %08x %08x %08x %08x %08x %08x %08x\n", sz, addr, line - &set->line[0], 
	//				src[0], src[1], src[2], src[3], src[4], src[5], src[6], src[7]);
				dptr = &line->dataW[(addr % CACHE_LINE_SIZE) / sizeof(uint32_t)];
				asm volatile(
					"	ldmia %0!, {r0, r1, r2, r3}	\n"
					"	stmia %1!, {r0, r1, r2, r3}	\n"
					"	ldmia %0!, {r0, r1, r2, r3}	\n"
					"	stmia %1!, {r0, r1, r2, r3}	\n"
					:"=l"(dummy1), "=l"(dummy2)
					:"0"(src), "1"(dptr)
					:"memory", "r0", "r1", "r2", "r3"
				);
				break;
	#endif
	#if CACHE_LINE_SIZE >= 64
			case 64:
	//			pr("write %u @ 0x%08x idx %u <- %08x %08x %08x %08x %08x %08x %08x %08x %08x %08x %08x %08x %08x %08x %08x %08x\n", sz, addr, line - &set->line[0], 
	//				src[0], src[1], src[2], src[3], src[4], src[5], src[6], src[7],
	//				src[8], src[9], src[10], src[11], src[12], src[13], src[14], src[15]);
				dptr = &line->dataW[(addr % CACHE_LINE_SIZE) / sizeof(uint32_t)];
				asm volatile(
					"	ldmia %0!, {r0, r1, r2, r3}	\n"
					"	stmia %1!, {r0, r1, r2, r3}	\n"
					"	ldmia %0!, {r0, r1, r2, r3}	\n"
					"	stmia %1!, {r0, r1, r2, r3}	\n"
					"	ldmia %0!, {r0, r1, r2, r3}	\n"
					"	stmia %1!, {r0, r1, r2, r3}	\n"
					:"=l"(dummy1), "=l"(dummy2)
					:"0"(src), "1"(dptr)
					:"memory", "r0", "r1", "r2", "r3"
				);
				break;
	#endif
			default:
				pr("unknown size %u\n", sz);
				while(1);
		}
		line->dirty = 1;
	}

#endif















//access code
//XXX: moving funcs to ram works well for more speed

#pragma GCC optimize ("Ofast")

static void __attribute__((naked)) spiRamPrvIobusRead32Bx4(uint32_t addr, void *dstP)
{
	asm volatile(
		".syntax unified				\n\t"
		"	push	{r4-r7}				\n\t"
		"	movs	r2, #0x60			\n\t"
		"	lsls	r2, #24				\n\t"	//r2 = PORT_IOBUS
		"	movs	r3, #1				\n\t"
		"	lsls	r3, #18				\n\t"	//r3 = nCS_bit
		"	mov		r12, r3				\n\t"	//r12 = nCS_bit
		"	str		r3, [r2, #0x14]		\n\t"	//nCS is low
		"	ldr		r4, =%0				\n\t"	//QSPI_ALL_PINS
		"	str		r4, [r2, #0x08]		\n\t"	//set all pins to out
		"	str		r4, [r2, #0x14]		\n\t"	//output low
		"	lsls	r3, #9				\n\t"	//r3 = CLK_bit
		"	str		r3, [r2, #0x18]		\n\t"	//clock high
		"	ldr		r5, =0x2cbbb		\n\t"	//command 0x0b
		"	str		r3, [r2, #0x14]		\n\t"	//clock low
		"	lsls	r0, #6				\n\t"	//left-justify address after dividing it by 4
		"	str		r5, [r2, #0x18]		\n\t"	//set command 2
		"	movs	r6, #6				\n\t"	//loop counter for address loop
		"	str		r3, [r2, #0x18]		\n\t"	//clock high
		"	ldr		r5, =0x4111			\n\t"	//broadcast mask
		"	str		r3, [r2, #0x14]		\n\t"	//clock low
		//send address
		"1:								\n\t"	//addr loop
		"	lsrs	r7, r0, #28			\n\t"
		"	muls	r7, r5				\n\t"
		"	str		r7, [r2, #0x18]		\n\t"	//set high bits for out
		"	eors	r7, r4				\n\t"
		"	str		r7, [r2, #0x14]		\n\t"	//set low bits for out
		"	str		r3, [r2, #0x18]		\n\t"	//clock high
		"	lsls	r0, #4				\n\t"
		"	subs	r6, #1				\n\t"
		"	str		r3, [r2, #0x14]		\n\t"	//clock low
		"	bne		1b					\n\t"
		
		"	str		r4, [r2, #0x04]		\n\t"	//set all pins to in
		
		//4 dummy cycles
		"	movs	r6, #4				\n\t"
		"1:								\n\t"
		"	str		r3, [r2, #0x18]		\n\t"	//clock high
		"	subs	r6, #1				\n\t"
		"	str		r3, [r2, #0x14]		\n\t"	//clock low
		"	bne		1b					\n\t"
		
		//recv data
		//r2 = base, r0 = free, r1 = dataPtr, r3 = CLK_bit, r4 = free(QSPI_ALL_PINS), r5 = free(0x4111), r6 = free(0), r7 = free
		"	movs	r6, #16				\n\t"
		"	ldr		r7, =0xfff			\n\t"
		"	subs	r1, #2				\n\t"
		"1:								\n\t"
		"	str		r3, [r2, #0x18]		\n\t"	//clock high
		"	adds	r1, #2				\n\t"
		"	ldr		r0, [r2, #0x20]		\n\t"
		"	lsrs	r4, r0, #14			\n\t"
		"	lsls	r4, r4, #12			\n\t"
		"	ands	r0, r7				\n\t"
		"	str		r3, [r2, #0x14]		\n\t"	//clock low
		"	adds	r0, r4				\n\t"
		"	strh	r0, [r1]			\n\t"
		"	subs	r6, #1				\n\t"
		"	bne		1b					\n\t"
		
		"	mov		r3, r12				\n\t"	//r3 = ncsMask
		"	pop		{r4-r7}				\n\t"
		"	str		r3, [r2, #0x18]		\n\t"
		"	bx		lr					\n\t"
		
		:
		:"i"(QSPI_ALL_PINS)
		:"cc"
	);
}

static void __attribute__((naked)) spiRamPrvIobusWrite32Bx4(uint32_t addr, const void *srcP)
{
	asm volatile(
		".syntax unified				\n\t"
		"	push	{r4-r7}				\n\t"
		"	movs	r2, #0x60			\n\t"
		"	lsls	r2, #24				\n\t"	//r2 = PORT_IOBUS
		"	movs	r3, #1				\n\t"
		"	lsls	r3, #18				\n\t"	//r3 = nCS_bit
		"	mov		r12, r3				\n\t"	//r12 = nCS_bit
		"	str		r3, [r2, #0x14]		\n\t"	//nCS is low
		"	ldr		r4, =%0				\n\t"	//QSPI_ALL_PINS
		"	str		r4, [r2, #0x08]		\n\t"	//set all pins to out
		"	str		r4, [r2, #0x14]		\n\t"	//output low
		"	lsls	r3, #9				\n\t"	//r3 = CLK_bit
		"	str		r3, [r2, #0x18]		\n\t"	//clock high
		"	ldr		r5, =0x8222			\n\t"	//command 2
		"	str		r3, [r2, #0x14]		\n\t"	//clock low
		"	lsls	r0, #6				\n\t"	//left-justify address (after dividing it by 4)
		"	str		r5, [r2, #0x18]		\n\t"	//set command 2
		"	movs	r6, #6				\n\t"	//loop counter for address loop
		"	str		r3, [r2, #0x18]		\n\t"	//clock high
		"	lsrs	r5, #1				\n\t"	//broadcast mask
		"	str		r3, [r2, #0x14]		\n\t"	//clock low
		//send address
		"1:								\n\t"	//addr loop
		"	lsrs	r7, r0, #28			\n\t"
		"	muls	r7, r5				\n\t"
		"	str		r7, [r2, #0x18]		\n\t"	//set high bits for out
		"	eors	r7, r4				\n\t"
		"	str		r7, [r2, #0x14]		\n\t"	//set low bits for out
		"	str		r3, [r2, #0x18]		\n\t"	//clock high
		"	lsls	r0, #4				\n\t"
		"	subs	r6, #1				\n\t"
		"	str		r3, [r2, #0x14]		\n\t"	//clock low
		"	bne		1b					\n\t"
		
		//send data
		//r2 = base, r0 = free, r1 = dataPtr, r3 = CLK_bit, r4 = QSPI_ALL_PINS, r5 = free(0x4111), r6 = free(0), r7 = free
		"	movs	r7, #16				\n\t"
		"	ldr		r6, =0xfff			\n\t"
		"1:								\n\t"
		"	ldrh	r0, [r1]			\n\t"
		"	lsrs	r5, r0, #12			\n\t"
		"	lsls	r5, #14				\n\t"
		"	ands	r0, r6				\n\t"
		"	adds	r0, r5				\n\t"
		"	str		r0, [r2, #0x18]		\n\t"	//set high bits for out
		"	eors	r0, r4				\n\t"
		"	str		r0, [r2, #0x14]		\n\t"	//set low bits for out
		"	adds	r1, #2				\n\t"
		"	str		r3, [r2, #0x18]		\n\t"	//clock high
		"	subs	r7, #1				\n\t"
		"	str		r3, [r2, #0x14]		\n\t"	//clock low
		"	bne		1b					\n\t"
		
		"	mov		r3, r12				\n\t"	//r3 = ncsMask
		"	pop		{r4-r7}				\n\t"
		"	str		r3, [r2, #0x18]		\n\t"
		"	bx		lr					\n\t"
		
		:
		:"i"(QSPI_ALL_PINS)
		:"cc"
	);
}

static void __attribute__((naked)) spiRamPrvIobusRead32Bx2(uint32_t addr, void *dstP)
{
	asm volatile(
		".syntax unified				\n\t"
		"	push	{r4-r7}				\n\t"
		"	movs	r2, #0x60			\n\t"
		"	lsls	r2, #24				\n\t"	//r2 = PORT_IOBUS
		"	movs	r3, #1				\n\t"
		"	lsls	r3, #18				\n\t"	//r3 = nCS_bit
		"	mov		r12, r3				\n\t"	//r12 = nCS_bit
		"	str		r3, [r2, #0x14]		\n\t"	//nCS is low
		"	ldr		r4, =%0				\n\t"	//QSPI_ALL_PINS
		"	str		r4, [r2, #0x08]		\n\t"	//set all pins to out
		"	str		r4, [r2, #0x14]		\n\t"	//output low
		"	lsls	r3, #9				\n\t"	//r3 = CLK_bit
		"	str		r3, [r2, #0x18]		\n\t"	//clock high
		"	movs	r5, 0xbb			\n\t"	//command 0x0b
		"	str		r3, [r2, #0x14]		\n\t"	//clock low
		"	lsls	r0, #7				\n\t"	//left-justify address after dividing it by 2
		"	str		r5, [r2, #0x18]		\n\t"	//set command 2
		"	movs	r6, #6				\n\t"	//loop counter for address loop
		"	str		r3, [r2, #0x18]		\n\t"	//clock high
		"	movs	r5, #0x11			\n\t"	//broadcast mask
		"	str		r3, [r2, #0x14]		\n\t"	//clock low
		//send address
		"1:								\n\t"	//addr loop
		"	lsrs	r7, r0, #28			\n\t"
		"	muls	r7, r5				\n\t"
		"	strb	r7, [r2, #0x10]		\n\t"	//set out
		"	str		r3, [r2, #0x18]		\n\t"	//clock high
		"	lsls	r0, #4				\n\t"
		"	subs	r6, #1				\n\t"
		"	str		r3, [r2, #0x14]		\n\t"	//clock low
		"	bne		1b					\n\t"
		
		"	str		r4, [r2, #0x04]		\n\t"	//set all pins to in
		
		//4 dummy cycles
		"	movs	r6, #4				\n\t"
		"1:								\n\t"
		"	str		r3, [r2, #0x18]		\n\t"	//clock high
		"	subs	r6, #1				\n\t"
		"	str		r3, [r2, #0x14]		\n\t"	//clock low
		"	bne		1b					\n\t"
		
		//recv data
		//r2 = base, r0 = free, r1 = dataPtr, r3 = CLK_bit, r4 = free(QSPI_ALL_PINS), r5 = free(0x4111), r6 = free(0), r7 = free
		"	movs	r6, #32				\n\t"
		"	subs	r1, #1				\n\t"
		"1:								\n\t"
		"	str		r3, [r2, #0x18]		\n\t"	//clock high
		"	adds	r1, #1				\n\t"
		"	ldr		r0, [r2, #0x20]		\n\t"
		"	strb	r0, [r1]			\n\t"
		"	str		r3, [r2, #0x14]		\n\t"	//clock low
		"	subs	r6, #1				\n\t"
		"	bne		1b					\n\t"
		
		"	mov		r3, r12				\n\t"	//r3 = ncsMask
		"	pop		{r4-r7}				\n\t"
		"	str		r3, [r2, #0x18]		\n\t"
		"	bx		lr					\n\t"
		
		:
		:"i"(QSPI_ALL_PINS)
		:"cc"
	);
}

static void __attribute__((naked)) spiRamPrvIobusWrite32Bx2(uint32_t addr, const void *srcP)
{
	asm volatile(
		".syntax unified				\n\t"
		"	push	{r4-r7}				\n\t"
		"	movs	r2, #0x60			\n\t"
		"	lsls	r2, #24				\n\t"	//r2 = PORT_IOBUS
		"	movs	r3, #1				\n\t"
		"	lsls	r3, #18				\n\t"	//r3 = nCS_bit
		"	mov		r12, r3				\n\t"	//r12 = nCS_bit
		"	str		r3, [r2, #0x14]		\n\t"	//nCS is low
		"	ldr		r4, =%0				\n\t"	//QSPI_ALL_PINS
		"	str		r4, [r2, #0x08]		\n\t"	//set all pins to out
		"	str		r4, [r2, #0x14]		\n\t"	//output low
		"	lsls	r3, #9				\n\t"	//r3 = CLK_bit
		"	str		r3, [r2, #0x18]		\n\t"	//clock high
		"	movs	r5, #0x22			\n\t"	//command 2
		"	str		r3, [r2, #0x14]		\n\t"	//clock low
		"	lsls	r0, #7				\n\t"	//left-justify address (after dividing it by 2)
		"	str		r5, [r2, #0x18]		\n\t"	//set command 2
		"	movs	r6, #6				\n\t"	//loop counter for address loop
		"	str		r3, [r2, #0x18]		\n\t"	//clock high
		"	lsrs	r5, #1				\n\t"	//broadcast mask
		"	str		r3, [r2, #0x14]		\n\t"	//clock low
		//send address
		"1:								\n\t"	//addr loop
		"	lsrs	r7, r0, #28			\n\t"
		"	muls	r7, r5				\n\t"
		"	strb	r7, [r2, #0x10]		\n\t"	//set out
		"	str		r3, [r2, #0x18]		\n\t"	//clock high
		"	lsls	r0, #4				\n\t"
		"	subs	r6, #1				\n\t"
		"	str		r3, [r2, #0x14]		\n\t"	//clock low
		"	bne		1b					\n\t"
		
		//send data
		//r2 = base, r0 = free, r1 = dataPtr, r3 = CLK_bit, r4 = QSPI_ALL_PINS, r5 = free(0x4111), r6 = free(0), r7 = free
		"	movs	r7, #32				\n\t"
		"1:								\n\t"
		"	ldrb	r0, [r1]			\n\t"
		"	strb	r0, [r2, #0x10]		\n\t"	//set out
		"	str		r3, [r2, #0x18]		\n\t"	//clock high
		"	adds	r1, #1				\n\t"
		"	subs	r7, #1				\n\t"
		"	str		r3, [r2, #0x14]		\n\t"	//clock low
		"	bne		1b					\n\t"
		
		"	mov		r3, r12				\n\t"	//r3 = ncsMask
		"	pop		{r4-r7}				\n\t"
		"	str		r3, [r2, #0x18]		\n\t"
		"	bx		lr					\n\t"
		
		:
		:"i"(QSPI_ALL_PINS)
		:"cc"
	);
}

static void __attribute__((naked)) spiRamPrvIobusRead32Bx1(uint32_t addr, void *dstP)
{
	asm volatile(
		".syntax unified				\n\t"
		"	push	{r4-r7}				\n\t"
		"	movs	r2, #0x60			\n\t"
		"	lsls	r2, #24				\n\t"	//r2 = PORT_IOBUS
		"	movs	r3, #1				\n\t"
		"	lsls	r3, #18				\n\t"	//r3 = nCS_bit
		"	mov		r12, r3				\n\t"	//r12 = nCS_bit
		"	str		r3, [r2, #0x14]		\n\t"	//nCS is low
		"	ldr		r4, =%0				\n\t"	//QSPI_ALL_PINS
		"	str		r4, [r2, #0x08]		\n\t"	//set all pins to out
		"	str		r4, [r2, #0x14]		\n\t"	//output low
		"	lsls	r3, #9				\n\t"	//r3 = CLK_bit
		"	str		r3, [r2, #0x18]		\n\t"	//clock high
		"	movs	r5, #0x0b			\n\t"	//command 0x0b
		"	str		r3, [r2, #0x14]		\n\t"	//clock low
		"	lsls	r0, #8				\n\t"	//left-justify address
		"	str		r5, [r2, #0x18]		\n\t"	//set command 2
		"	movs	r6, #6				\n\t"	//loop counter for address loop
		"	str		r3, [r2, #0x18]		\n\t"	//clock high
		"	str		r3, [r2, #0x14]		\n\t"	//clock low
		//send address
		"1:								\n\t"	//addr loop
		"	lsrs	r7, r0, #28			\n\t"
		"	strb	r7, [r2, #0x10]		\n\t"	//set out
		"	str		r3, [r2, #0x18]		\n\t"	//clock high
		"	lsls	r0, #4				\n\t"
		"	subs	r6, #1				\n\t"
		"	str		r3, [r2, #0x14]		\n\t"	//clock low
		"	bne		1b					\n\t"
		
		"	str		r4, [r2, #0x04]		\n\t"	//set all pins to in
		
		//4 dummy cycles
		"	movs	r6, #4				\n\t"
		"1:								\n\t"
		"	str		r3, [r2, #0x18]		\n\t"	//clock high
		"	subs	r6, #1				\n\t"
		"	str		r3, [r2, #0x14]		\n\t"	//clock low
		"	bne		1b					\n\t"
		
		//recv data
		//r2 = base, r0 = free, r1 = dataPtr, r3 = CLK_bit, r4 = free(QSPI_ALL_PINS), r5 = free(0x4111), r6 = free(0), r7 = free
		"	movs	r6, #32				\n\t"
		"	movs	r5, #0x0f			\n\t"
		"	subs	r1, #1				\n\t"
		"1:								\n\t"
		"	str		r3, [r2, #0x18]		\n\t"	//clock high
		"	ldr		r0, [r2, #0x20]		\n\t"	//load
		"	lsls	r0, #4				\n\t"
		"	str		r3, [r2, #0x14]		\n\t"	//clock low
		"	adds	r1, #1				\n\t"
		"	str		r3, [r2, #0x18]		\n\t"	//clock high
		"	dsb sy						\n\t"
		"	ldr		r4, [r2, #0x20]		\n\t"	//load
		"	ands	r4, r5				\n\t"
		"	orrs	r0, r4				\n\t"
		"	str		r3, [r2, #0x14]		\n\t"	//clock low
		"	strb	r0, [r1]			\n\t"
		"	subs	r6, #1				\n\t"
		"	bne		1b					\n\t"
		
		"	mov		r3, r12				\n\t"	//r3 = ncsMask
		"	pop		{r4-r7}				\n\t"
		"	str		r3, [r2, #0x18]		\n\t"
		"	bx		lr					\n\t"
		
		:
		:"i"(QSPI_ALL_PINS)
		:"cc"
	);
}

static void __attribute__((naked)) spiRamPrvIobusWrite32Bx1(uint32_t addr, const void *srcP)
{
	asm volatile(
		".syntax unified				\n\t"
		"	push	{r4-r7}				\n\t"
		"	movs	r2, #0x60			\n\t"
		"	lsls	r2, #24				\n\t"	//r2 = PORT_IOBUS
		"	movs	r3, #1				\n\t"
		"	lsls	r3, #18				\n\t"	//r3 = nCS_bit
		"	mov		r12, r3				\n\t"	//r12 = nCS_bit
		"	str		r3, [r2, #0x14]		\n\t"	//nCS is low
		"	ldr		r4, =%0				\n\t"	//QSPI_ALL_PINS
		"	str		r4, [r2, #0x08]		\n\t"	//set all pins to out
		"	str		r4, [r2, #0x14]		\n\t"	//output low
		"	lsls	r3, #9				\n\t"	//r3 = CLK_bit
		"	str		r3, [r2, #0x18]		\n\t"	//clock high
		"	movs	r5, #0x02			\n\t"	//command 2
		"	str		r3, [r2, #0x14]		\n\t"	//clock low
		"	lsls	r0, #8				\n\t"	//left-justify address
		"	str		r5, [r2, #0x18]		\n\t"	//set command 2
		"	movs	r6, #6				\n\t"	//loop counter for address loop
		"	str		r3, [r2, #0x18]		\n\t"	//clock high
		"	str		r3, [r2, #0x14]		\n\t"	//clock low
		//send address
		"1:								\n\t"	//addr loop
		"	lsrs	r7, r0, #28			\n\t"
		"	strb	r7, [r2, #0x10]		\n\t"	//set out
		"	str		r3, [r2, #0x18]		\n\t"	//clock high
		"	lsls	r0, #4				\n\t"
		"	subs	r6, #1				\n\t"
		"	str		r3, [r2, #0x14]		\n\t"	//clock low
		"	bne		1b					\n\t"
		
		//send data
		//r2 = base, r0 = free, r1 = dataPtr, r3 = CLK_bit, r4 = QSPI_ALL_PINS, r5 = free(0x4111), r6 = free(0), r7 = free
		"	movs	r7, #32				\n\t"
		"1:								\n\t"
		"	ldrb	r0, [r1]			\n\t"
		"	lsrs	r6, r0, #4			\n\t"
		"	strb	r6, [r2, #0x10]		\n\t"	//set out
		"	str		r3, [r2, #0x18]		\n\t"	//clock high
		"	adds	r1, #1				\n\t"
		"	str		r3, [r2, #0x14]		\n\t"	//clock low
		"	strb	r0, [r2, #0x10]		\n\t"	//set out
		"	str		r3, [r2, #0x18]		\n\t"	//clock high
		"	subs	r7, #1				\n\t"
		"	str		r3, [r2, #0x14]		\n\t"	//clock low
		"	bne		1b					\n\t"
		
		"	mov		r3, r12				\n\t"	//r3 = ncsMask
		"	pop		{r4-r7}				\n\t"
		"	str		r3, [r2, #0x18]		\n\t"
		"	bx		lr					\n\t"
		
		:
		:"i"(QSPI_ALL_PINS)
		:"cc"
	);
}

	#define SENDALL1(_r)							\
		"	str		" _r ", [r3, #0x28]		\n\t"
	#define SENDALL2(_r)							\
		SENDALL1(_r)								\
		"	str		" _r ", [r5, #0x28]		\n\t"
	#define SENDALL4(_r)							\
		SENDALL2(_r)								\
		"	str		" _r ", [r4, #0x28]		\n\t"	\
		"	str		" _r ", [r6, #0x28]		\n\t"
	#define WAIT(_r, _sh, _u)						\
		"1:									\n\t"	\
		"	ldrb	" _r ", [" _u ", #0x18]	\n\t"	\
		"	lsls	" _r ", #" _sh "		\n\t"	\
		"	bpl		1b						\n\t"
	#define WAIT_DRE4(_r)							\
		WAIT(_r, "31", "r6")
	#define WAIT_TXC4(_r)							\
		WAIT(_r, "30", "r6")
	#define WAIT_RXC4(_r)							\
		WAIT(_r, "29", "r6")
	#define WAIT_DRE2(_r)							\
		WAIT(_r, "31", "r5")
	#define WAIT_TXC2(_r)							\
		WAIT(_r, "30", "r5")
	#define WAIT_RXC2(_r)							\
		WAIT(_r, "29", "r5")
	#define WAIT_DRE1(_r)							\
		WAIT(_r, "31", "r3")
	#define WAIT_TXC1(_r)							\
		WAIT(_r, "30", "r3")
	#define WAIT_RXC1(_r)							\
		WAIT(_r, "29", "r3")
	#define CONSUME1(_r)							\
		"	ldr		" _r ", [r3, #0x28]		\n\t"
	#define CONSUME2(_r)							\
		CONSUME1(_r)								\
		"	ldr		" _r ", [r5, #0x28]		\n\t"
	#define CONSUME4(_r)							\
		CONSUME2(_r)								\
		"	ldr		" _r ", [r4, #0x28]		\n\t"	\
		"	ldr		" _r ", [r6, #0x28]		\n\t"
	#define READ1(_r, _f)							\
		"	ldr		" _r ", [r3, #0x28]		\n\t"	\
		"	strb	" _r ", [r1, #0 + " _f "]\n\t"
	#define READ2(_r, _f)							\
		READ1(_r, _f)								\
		"	ldr		" _r ", [r5, #0x28]		\n\t"	\
		"	strb	" _r ", [r1, #1 + " _f "]\n\t"
	#define READ4(_r, _f)							\
		READ2(_r, _f)								\
		"	ldr		" _r ", [r4, #0x28]		\n\t"	\
		"	strb	" _r ", [r1, #2 + " _f "]\n\t"	\
		"	ldr		" _r ", [r6, #0x28]		\n\t"	\
		"	strb	" _r ", [r1, #3 + " _f "]\n\t"
	#define WRITE1(_r, _f)							\
		"	ldrb	" _r ", [r1, #0 + " _f "]\n\t"	\
		"	str		" _r ", [r3, #0x28]		\n\t"
	#define WRITE2(_r, _f)							\
		WRITE1(_r, _f)								\
		"	ldrb	" _r ", [r1, #1 + " _f "]\n\t"	\
		"	str		" _r ", [r5, #0x28]		\n\t"
	#define WRITE4(_r, _f)							\
		WRITE2(_r, _f)								\
		"	ldrb	" _r ", [r1, #2 + " _f "]\n\t"	\
		"	str		" _r ", [r4, #0x28]		\n\t"	\
		"	ldrb	" _r ", [r1, #3 + " _f "]\n\t"	\
		"	str		" _r ", [r6, #0x28]		\n\t"
	#define WAITCONSUME4(_r)						\
		WAIT_DRE4(_r)								\
		CONSUME4(_r)
	#define SEND_READ4(_s, _r, _f)					\
		SENDALL4(_s)								\
		READ4(_r, _f)
	#define WAIT_WRITE4(_r, _f)						\
		WAIT_DRE4(_r)								\
		WRITE4(_r, _f)
	#define WAITCONSUME2(_r)						\
		WAIT_DRE2(_r)								\
		CONSUME2(_r)
	#define SEND_READ2(_s, _r, _f)					\
		SENDALL2(_s)								\
		READ2(_r, _f)
	#define WAIT_WRITE2(_r, _f)						\
		WAIT_DRE2(_r)								\
		WRITE2(_r, _f)
	#define WAITCONSUME1(_r)						\
		WAIT_DRE1(_r)								\
		CONSUME1(_r)
	#define SEND_READ1(_s, _r, _f)					\
		SENDALL1(_s)								\
		READ1(_r, _f)
	#define WAIT_WRITE1(_r, _f)						\
		WAIT_DRE1(_r)								\
		WRITE1(_r, _f)
		

//dma can make the one-chip and in some cases two-chip version of this faster. i don't care. v1.3 boards are better anyways :)
static void __attribute__((naked)) spiRamPrvSercomRead32Bx4(uint32_t addr, void *dst)
{
	asm volatile(
		"	.syntax unified				\n\t"
		"	lsrs	r0, #2				\n\t"
		"	push	{r4-r6}				\n\t"
		"	movs	r4, #0x60			\n\t"	//R4 = PORT_IOBUS->Group[0]
		"	lsls	r4, #24				\n\t"
		"	movs	r5, #1				\n\t"
		"	lsls	r5, #14				\n\t"
		"	str		r5, [r4, #0x14]		\n\t"	//PORT->Group[0].OUTCLR.reg = PORT_PA14
		"	mov		r12, r5				\n\t"
		"	ldr		r3, =%0				\n\t"	//R3 = sercom0
		"	ldr		r6, =%1				\n\t"
		"	adds	r4, r3, r6			\n\t"	//R4 = sercom1
		"	adds	r5, r4, r6			\n\t"	//R5 = sercom2
		"	adds	r6, r5, r6			\n\t"	//R6 = sercom3
		"	movs	r2, #0x03			\n\t"
		"	cpsid	i					\n\t"
		SENDALL4("r2")
		"	lsrs	r2, r0, #16			\n\t"
		SENDALL4("r2")
		WAITCONSUME4("r2")
		"	lsrs	r2, r0, #8			\n\t"
		SENDALL4("r2")
		WAITCONSUME4("r2")
		SENDALL4("r0")
		WAITCONSUME4("r2")
		"	movs	r0, #1				\n\t"	//low byte is low, high we'll use later
		"	lsls	r0, #14				\n\t"
		SENDALL4("r0")
		WAITCONSUME4("r2")
		WAIT_RXC4("r2")
		
		SEND_READ4("r0", "r2", "0x00")
		SEND_READ4("r0", "r2", "0x04")
		SEND_READ4("r0", "r2", "0x08")
		SEND_READ4("r0", "r2", "0x0c")
		SEND_READ4("r0", "r2", "0x10")
		SEND_READ4("r0", "r2", "0x14")
		SEND_READ4("r0", "r2", "0x18")
		WAIT_RXC4("r2")
		READ4("r2", "0x1c")
		
		"	cpsie	i					\n\t"
		"	mov		r0, r12				\n\t"
		"	movs	r4, #0x60			\n\t"	//R4 = PORT_IOBUS->Group[0]
		"	lsls	r4, #24				\n\t"
		"	str		r0, [r4, #0x18]		\n\t"	//PORT->Group[0].OUTSET.reg = PORT_PA14
		"	pop		{r4-r6}				\n\t"
		"	bx		lr					\n\t"
	::"i"(SERCOM0), "i"(((uint32_t)SERCOM1) - ((uint32_t)SERCOM0)):"memory");
}

static void __attribute__((naked)) spiRamPrvSercomWrite32Bx4(uint32_t addr, const void *src)
{
	asm volatile(
		"	.syntax unified				\n\t"
		"	lsrs	r0, #2				\n\t"
		"	push	{r4-r6}				\n\t"
		"	movs	r4, #0x60			\n\t"	//R4 = PORT_IOBUS->Group[0]
		"	lsls	r4, #24				\n\t"
		"	movs	r5, #1				\n\t"
		"	lsls	r5, #14				\n\t"
		"	str		r5, [r4, #0x14]		\n\t"	//PORT->Group[0].OUTCLR.reg = PORT_PA14
		"	mov		r12, r5				\n\t"
		"	ldr		r3, =%0				\n\t"	//R3 = sercom0
		"	ldr		r6, =%1				\n\t"
		"	adds	r4, r3, r6			\n\t"	//R4 = sercom1
		"	adds	r5, r4, r6			\n\t"	//R5 = sercom2
		"	adds	r6, r5, r6			\n\t"	//R6 = sercom3
		"	movs	r2, #0x02			\n\t"
		"	cpsid	i					\n\t"
		SENDALL4("r2")
		"	lsrs	r2, r0, #16			\n\t"
		SENDALL4("r2")
		WAIT_DRE4("r2")
		"	lsrs	r2, r0, #8			\n\t"
		SENDALL4("r2")
		WAIT_DRE4("r2")
		SENDALL4("r0")
		
		WAIT_WRITE4("r0", "0x00")
		WAIT_WRITE4("r0", "0x04")
		WAIT_WRITE4("r0", "0x08")
		WAIT_WRITE4("r0", "0x0c")
		WAIT_WRITE4("r0", "0x10")
		WAIT_WRITE4("r0", "0x14")
		WAIT_WRITE4("r0", "0x18")
		WAIT_WRITE4("r0", "0x1c")
		WAIT_TXC4("r0")
		CONSUME4("r0")
		CONSUME4("r0")
		CONSUME4("r0")
		
		"	cpsie	i					\n\t"
		"	mov		r0, r12				\n\t"
		"	movs	r4, #0x60			\n\t"	//R4 = PORT_IOBUS->Group[0]
		"	lsls	r4, #24				\n\t"
		"	str		r0, [r4, #0x18]		\n\t"	//PORT->Group[0].OUTSET.reg = PORT_PA14
		"	pop		{r4-r6}				\n\t"
		"	bx		lr					\n\t"
	::"i"(SERCOM0), "i"(((uint32_t)SERCOM1) - ((uint32_t)SERCOM0)):"memory");	
}

//this could be done in C, but i am too lazy, just buy 4 RAM chips :D
static void spiRamPrvSercomRead32Bx2(uint32_t addr, void *dstP)
{
	uint8_t *dst = (uint8_t*)dstP;
	uint_fast8_t i;
	
	addr = 0x03000000 + (addr >> 1);
	
	PORT_IOBUS->Group[0].OUTCLR.reg = PORT_PA14;
	for (i = 0; i < 4; i++, addr <<= 8) {
		
		while (!SERCOM2->SPI.INTFLAG.bit.DRE);
		SERCOM0->SPI.DATA.reg = addr >> 24;
		SERCOM2->SPI.DATA.reg = addr >> 24;
	}
	while (!SERCOM2->SPI.INTFLAG.bit.TXC);
	while (!SERCOM2->SPI.INTFLAG.bit.RXC);
	(void)SERCOM0->SPI.DATA.reg;
	(void)SERCOM2->SPI.DATA.reg;
	(void)SERCOM0->SPI.DATA.reg;
	(void)SERCOM2->SPI.DATA.reg;
	(void)SERCOM0->SPI.DATA.reg;
	(void)SERCOM2->SPI.DATA.reg;
	for (i = 0; i < 32; i += 2, dst += 2) {
		
		SERCOM0->SPI.DATA.reg = 0;
		SERCOM2->SPI.DATA.reg = 0;
		while (!SERCOM2->SPI.INTFLAG.bit.RXC);
		dst[0] = SERCOM0->SPI.DATA.reg;
		dst[1] = SERCOM2->SPI.DATA.reg;
	}
	while (!SERCOM2->SPI.INTFLAG.bit.TXC);
	asm volatile("dsb sy");	//give it a sec cause the flag goes up before we're really done
	PORT_IOBUS->Group[0].OUTSET.reg = PORT_PA14;
}

//this could be done in C, but i am too lazy, just buy 4 RAM chips :D
static void spiRamPrvSercomWrite32Bx2(uint32_t addr, const void *srcP)
{
	const uint8_t *src = (const uint8_t*)srcP;
	uint_fast8_t i;
	
	addr = 0x02000000 + (addr >> 1);
	
	PORT_IOBUS->Group[0].OUTCLR.reg = PORT_PA14;
	for (i = 0; i < 4; i++, addr <<= 8) {
		
		while (!SERCOM2->SPI.INTFLAG.bit.DRE);
		SERCOM0->SPI.DATA.reg = addr >> 24;
		SERCOM2->SPI.DATA.reg = addr >> 24;
	}
	for (i = 0; i < 32; i += 2, src += 2) {
		
		while (!SERCOM2->SPI.INTFLAG.bit.DRE);
		SERCOM0->SPI.DATA.reg = src[0];
		SERCOM2->SPI.DATA.reg = src[1];
	}
	while (!SERCOM2->SPI.INTFLAG.bit.TXC);
	while (!SERCOM2->SPI.INTFLAG.bit.RXC);
	(void)SERCOM0->SPI.DATA.reg;
	(void)SERCOM2->SPI.DATA.reg;
	(void)SERCOM0->SPI.DATA.reg;
	(void)SERCOM2->SPI.DATA.reg;
	(void)SERCOM0->SPI.DATA.reg;
	(void)SERCOM2->SPI.DATA.reg;
	PORT_IOBUS->Group[0].OUTSET.reg = PORT_PA14;
}


//this could be done in C, but i am too lazy, just buy 4 RAM chips :D
static void spiRamPrvSercomRead32Bx1(uint32_t addr, void *dstP)
{
	uint8_t *dst = (uint8_t*)dstP;
	uint_fast8_t i;
	
	addr = 0x03000000 + addr;
	
	PORT_IOBUS->Group[0].OUTCLR.reg = PORT_PA14;
	for (i = 0; i < 4; i++, addr <<= 8) {
		
		while (!SERCOM0->SPI.INTFLAG.bit.DRE);
		SERCOM0->SPI.DATA.reg = addr >> 24;
	}
	while (!SERCOM0->SPI.INTFLAG.bit.TXC);
	while (!SERCOM0->SPI.INTFLAG.bit.RXC);
	(void)SERCOM0->SPI.DATA.reg;
	(void)SERCOM0->SPI.DATA.reg;
	(void)SERCOM0->SPI.DATA.reg;
	for (i = 0; i < 32; i++) {
		
		SERCOM0->SPI.DATA.reg = 0;
		while (!SERCOM0->SPI.INTFLAG.bit.RXC);
		*dst++ = SERCOM0->SPI.DATA.reg;
	}
	while (!SERCOM0->SPI.INTFLAG.bit.TXC);
	asm volatile("dsb sy");	//give it a sec cause the flag goes up before we're really done
	PORT_IOBUS->Group[0].OUTSET.reg = PORT_PA14;
}

//this could be done in C, but i am too lazy, just buy 4 RAM chips :D
static void spiRamPrvSercomWrite32Bx1(uint32_t addr, const void *srcP)
{
	const uint8_t *src = (const uint8_t*)srcP;
	uint_fast8_t i;
	
	addr = 0x02000000 + addr;
	
	PORT_IOBUS->Group[0].OUTCLR.reg = PORT_PA14;
	for (i = 0; i < 4; i++, addr <<= 8) {
		
		while (!SERCOM0->SPI.INTFLAG.bit.DRE);
		SERCOM0->SPI.DATA.reg = addr >> 24;
	}
	for (i = 0; i < 32; i++) {
		
		while (!SERCOM0->SPI.INTFLAG.bit.DRE);
		SERCOM0->SPI.DATA.reg = *src++;
	}
	while (!SERCOM0->SPI.INTFLAG.bit.TXC);
	while (!SERCOM0->SPI.INTFLAG.bit.RXC);
	(void)SERCOM0->SPI.DATA.reg;
	(void)SERCOM0->SPI.DATA.reg;
	(void)SERCOM0->SPI.DATA.reg;
	PORT_IOBUS->Group[0].OUTSET.reg = PORT_PA14;
}

	#undef SENDALL4
	#undef WAIT4
	#undef WAIT_DRE4
	#undef WAIT_TXC4
	#undef WAIT_RXC4
	#undef CONSUME4
	#undef READ4
	#undef WRITE4
	#undef WAITCONSUME4
	#undef SEND_READ4
	#undef WAIT_WRITE4










//init code

#pragma GCC optimize ("Os")

static void spiRamPrvClockFirst(void)
{
	asm volatile("dsb sy\ndsb sy\n");
	PORT_IOBUS->Group[0].OUTSET.reg = PORT_PA27;
	asm volatile("dsb sy\ndsb sy\n");
}

static void spiRamPrvClockSecond(void)
{
	asm volatile("dsb sy\ndsb sy\n");
	PORT_IOBUS->Group[0].OUTCLR.reg = PORT_PA27;
	asm volatile("dsb sy\ndsb sy\n");
}

static void spiRamPrvSingleClock(void)
{
	spiRamPrvClockFirst();
	spiRamPrvClockSecond();
}

static void spiRamPrvQspiNibbleOut(uint_fast8_t val)
{
	uint32_t outSet = val & 0x0f, outClr = outSet ^ 0x0f;
	
	outSet = outSet + (outSet << 4) + (outSet << 8) + (outSet << 14);
	outClr = outClr + (outClr << 4) + (outClr << 8) + (outClr << 14);
		
	PORT_IOBUS->Group[0].OUTSET.reg = outSet;
	PORT_IOBUS->Group[0].OUTCLR.reg = outClr;
	spiRamPrvSingleClock();
}

static void spiRamPrvQspiByteOut(uint_fast8_t val)
{
	spiRamPrvQspiNibbleOut(val >> 4);
	spiRamPrvQspiNibbleOut(val);
}

static void spiRamPrvSpiByteV13(uint8_t *out, uint_fast8_t val)
{
	static const uint8_t pinsIn[] = {PIN_PA01, PIN_PA05, PIN_PA09, PIN_PA15, };
	uint_fast8_t i, j;
	uint32_t inVal;
	
	for (i = 0; i < 8; i++, val <<= 1) {
		
		if (val & 0x80)
			PORT_IOBUS->Group[0].OUTSET.reg = QSPI_ALL_MOSI;
		else
			PORT_IOBUS->Group[0].OUTCLR.reg = QSPI_ALL_MOSI;
		spiRamPrvClockFirst();
		inVal = PORT_IOBUS->Group[0].IN.reg;
		spiRamPrvClockSecond();
		
		for (j = 0; j < MAX_NUM_CHIPS; j++) {
			
			uint_fast8_t v = (inVal >> pinsIn[j]) & 1;
			
			out[j] = out[j] * 2 + v;
		}
	}
}

static void spiRamPrvSpiByteV12(uint8_t *out, uint_fast8_t val)
{
	static Sercom  * const mSercoms[] = {SERCOM0, SERCOM1, SERCOM2, SERCOM3, };
	uint_fast8_t i;
	
	for (i = 0; i < MAX_NUM_CHIPS; i++) {
		
		while (!mSercoms[i]->SPI.INTFLAG.bit.DRE);
		mSercoms[i]->SPI.DATA.reg = val;
		while (!mSercoms[i]->SPI.INTFLAG.bit.RXC);
		*out++ = mSercoms[i]->SPI.DATA.reg;
	}
}

static void spiRamPrvSelect(void)
{
	PORT_IOBUS->Group[0].OUTCLR.reg = PORT_PA18;
}

static void spiRamPrvDeselect(void)
{
	PORT_IOBUS->Group[0].OUTSET.reg = PORT_PA18;
}

static void spiRamPrvQspiOutCfg(void)
{
	//all pins out
	PORT_IOBUS->Group[0].DIRSET.reg = QSPI_ALL_PINS;
}

static void spiRamPrvQspiInCfg(void)
{
	//all pins in
	PORT_IOBUS->Group[0].DIRCLR.reg = QSPI_ALL_PINS;
}

static void spiRamPrvSpiCfg(void)
{
	uint32_t outPins = QSPI_ALL_MOSI;
	
	//all except MOSI in, MOSI is out
	PORT_IOBUS->Group[0].DIRSET.reg = outPins;
	PORT_IOBUS->Group[0].DIRCLR.reg = QSPI_ALL_PINS &~ outPins;
}

static void spiRamPrvInitV13(uint8_t *id, const MemRead32Bfunc **readFuncsP, const MemWrite32Bfunc **writeFuncsP)
{
	static const MemWrite32Bfunc mWriteFuncsV13[] = {spiRamPrvIobusWrite32Bx1, spiRamPrvIobusWrite32Bx2, spiRamPrvIobusWrite32Bx4};
	static const MemRead32Bfunc mReadFuncsV13[] = {spiRamPrvIobusRead32Bx1, spiRamPrvIobusRead32Bx2, spiRamPrvIobusRead32Bx4};
	uint8_t dummy[MAX_NUM_CHIPS];
	uint_fast8_t i;
	
	*readFuncsP = mReadFuncsV13;
	*writeFuncsP = mWriteFuncsV13;
	
	//configure clock and nCS pins. clock low, nCS high
	PORT_IOBUS->Group[0].DIRSET.reg = PORT_PA27 | PORT_PA18;
	PORT_IOBUS->Group[0].OUTSET.reg = PORT_PA18;
	PORT_IOBUS->Group[0].OUTCLR.reg = PORT_PA27;
	
	for (i = 0; i < 19; i++)
		PORT->Group[0].PINCFG[i].reg = PORT_PINCFG_DRVSTR | PORT_PINCFG_INEN | PORT_PINCFG_PULLEN;
	PORT->Group[0].PINCFG[27].reg = PORT_PINCFG_DRVSTR | PORT_PINCFG_PULLEN;
		
	//we do not know what state we're in, so send "enter qspi" code in spi mode and then "exit qspi" code in qspi mode
	spiRamPrvSpiCfg();
	spiRamPrvSelect();
	spiRamPrvSpiByteV13(dummy, 0x35);
	spiRamPrvDeselect();
	
	spiRamPrvQspiOutCfg();
	spiRamPrvSelect();
	spiRamPrvQspiByteOut(0xf5);
	spiRamPrvDeselect();
	
	pr("query\n");
	
	//now we can query infos
	spiRamPrvSpiCfg();
	spiRamPrvSelect();
	for (i = 0; i < 4; i++)
		spiRamPrvSpiByteV13(dummy, i ? 0x00 : 0x9f);
	for (i = 0; i < SIZEOF_ID; i++)
		spiRamPrvSpiByteV13(id + i * MAX_NUM_CHIPS, 0x00);
	spiRamPrvDeselect();
	
	//go to qspi mode
	spiRamPrvSpiCfg();
	spiRamPrvSelect();
	spiRamPrvSpiByteV13(dummy, 0x35);
	spiRamPrvDeselect();
}

static void spiRamPrvInitV12(uint8_t *id, const MemRead32Bfunc **readFuncsP, const MemWrite32Bfunc **writeFuncsP)
{
	static const MemWrite32Bfunc mWriteFuncsV12[] = {spiRamPrvSercomWrite32Bx1, spiRamPrvSercomWrite32Bx2, spiRamPrvSercomWrite32Bx4};
	static const MemRead32Bfunc mReadFuncsV12[] = {spiRamPrvSercomRead32Bx1, spiRamPrvSercomRead32Bx2, spiRamPrvSercomRead32Bx4};
	uint_fast8_t i;
	
	*readFuncsP = mReadFuncsV12;
	*writeFuncsP = mWriteFuncsV12;
	
	//dirs
	PORT->Group[0].DIRSET.reg =	PORT_PA14 |					//nCS for ram
								PORT_PA04 | PORT_PA05 |		//SERCOM0
								PORT_PA16 | PORT_PA17 |		//SERCOM1
								PORT_PA08 | PORT_PA09;		//SERCOM2
	PORT->Group[0].DIRCLR.reg = PORT_PA15 |					//CARDDET
								PORT_PA06 |					//SERCOM0
								PORT_PA18 |					//SERCOM1
								PORT_PA10;					//SERCOM2

	//function outputs
	PORT->Group[0].PINCFG[PIN_PA04].reg = PORT_PINCFG_DRVSTR | PORT_PINCFG_PMUXEN;
	PORT->Group[0].PINCFG[PIN_PA05].reg = PORT_PINCFG_DRVSTR | PORT_PINCFG_PMUXEN;
	PORT->Group[0].PINCFG[PIN_PA08].reg = PORT_PINCFG_DRVSTR | PORT_PINCFG_PMUXEN;
	PORT->Group[0].PINCFG[PIN_PA09].reg = PORT_PINCFG_DRVSTR | PORT_PINCFG_PMUXEN;
	PORT->Group[0].PINCFG[PIN_PA16].reg = PORT_PINCFG_DRVSTR | PORT_PINCFG_PMUXEN;
	PORT->Group[0].PINCFG[PIN_PA17].reg = PORT_PINCFG_DRVSTR | PORT_PINCFG_PMUXEN;

	//function inputs
	PORT->Group[0].PINCFG[PIN_PA06].reg = PORT_PINCFG_INEN | PORT_PINCFG_PMUXEN | PORT_PINCFG_PULLEN;
	PORT->Group[0].PINCFG[PIN_PA10].reg = PORT_PINCFG_INEN | PORT_PINCFG_PMUXEN | PORT_PINCFG_PULLEN;
	PORT->Group[0].PINCFG[PIN_PA18].reg = PORT_PINCFG_INEN | PORT_PINCFG_PMUXEN | PORT_PINCFG_PULLEN;
	
	//gpio outputs
	PORT->Group[0].PINCFG[PIN_PA14].reg = PORT_PINCFG_INEN | PORT_PINCFG_DRVSTR;
	
	//gpio inputs
	PORT->Group[0].PINCFG[PIN_PA15].reg = PORT_PINCFG_INEN | PORT_PINCFG_PULLEN;
	
	//pull and output values
	PORT->Group[0].OUTSET.reg = PORT_PA15 | PORT_PA14 |  PORT_PA06 | PORT_PA10 | PORT_PA18;
	
	//functions
	SET_MUX(PIN_PA04, 3);
	SET_MUX(PIN_PA05, 3);
	SET_MUX(PIN_PA06, 3);
	
	SET_MUX(PIN_PA08, 3);
	SET_MUX(PIN_PA09, 3);
	SET_MUX(PIN_PA10, 3);
	
	SET_MUX(PIN_PA16, 2);
	SET_MUX(PIN_PA17, 2);
	SET_MUX(PIN_PA18, 2);
	
	//read IDs
	
	//nCS low
	PORT_IOBUS->Group[0].OUTCLR.reg = PORT_PA14;
	
	//send command and addrs
	spiRamPrvSpiByteV12(id, 0x9f);
	for (i = 0; i < 3; i++)
		spiRamPrvSpiByteV12(id, 0x00);
	
	//get data
	for (i = 0; i < SIZEOF_ID; i++)
		spiRamPrvSpiByteV12(id + MAX_NUM_CHIPS * i, 0x00);
	
	//nCS high
	PORT_IOBUS->Group[0].OUTSET.reg = PORT_PA14;
}

static uint_fast8_t spiRamPrvBbSpiV12(uint_fast8_t val)	//only to ram #0
{
	uint_fast8_t i;
	
	for (i = 0; i < 8; i++) {
		
		if (val & 0x80)
			PORT_IOBUS->Group[0].OUTSET.reg = PORT_PA04;
		else
			PORT_IOBUS->Group[0].OUTCLR.reg = PORT_PA04;
		asm volatile("dsb sy");
		PORT_IOBUS->Group[0].OUTSET.reg = PORT_PA05;
		asm volatile("dsb sy");
		val <<= 1;
		if (PORT_IOBUS->Group[0].IN.reg & PORT_PA06)
			val++;
		PORT_IOBUS->Group[0].OUTCLR.reg = PORT_PA05;
	}
	
	return val;
}

static bool spiRamPrvIsPreV13board(uint8_t *id)
{
	#define NUM_TIMES		4
	
	uint_fast8_t i, j;
	
	//do not bother with SERCOM for probe - use BBSPI
	
	//dir
	PORT_IOBUS->Group[0].DIRSET.reg = PORT_PA04 | PORT_PA05 | PORT_PA14;
	PORT_IOBUS->Group[0].DIRCLR.reg = PORT_PA06;
	
	//cfg
	PORT->Group[0].PINCFG[PIN_PA04].reg = PORT_PINCFG_DRVSTR;
	PORT->Group[0].PINCFG[PIN_PA05].reg = PORT_PINCFG_DRVSTR;
	PORT->Group[0].PINCFG[PIN_PA06].reg = PORT_PINCFG_INEN | PORT_PINCFG_PULLEN;
	PORT->Group[0].PINCFG[PIN_PA14].reg = PORT_PINCFG_DRVSTR;
	
	//vals
	PORT_IOBUS->Group[0].OUTSET.reg = PORT_PA14;
	PORT_IOBUS->Group[0].OUTCLR.reg = PORT_PA04 | PORT_PA05 | PORT_PA06;
	
	
	//now let's try an ID command a few times and verify we get not 0xFF, not 0x00, and always same
	for (i = 0; i < NUM_TIMES; i++) {
		
		uint_fast8_t and = 0xff, orr = 0x00, t;
		bool fail = false;
		
		//chip select
		PORT_IOBUS->Group[0].OUTCLR.reg = PORT_PA14;
		
		//command
		spiRamPrvBbSpiV12(0x9f);
		
		//address
		for (j = 0; j < 3; j++)
			spiRamPrvBbSpiV12(0x00);
		
		//data
		for (j = 0; j < SIZEOF_ID; j++) {
			
			t = spiRamPrvBbSpiV12(0x00);
			
			and &= t;
			orr |= t;
			
			if (i == 0)
				id[j] = t;
			else if (id[j] != t)
				fail = true;
		}
		
		//chip deselect
		PORT_IOBUS->Group[0].OUTSET.reg = PORT_PA14;
		
		if (and == 0xff || orr == 0x00 || fail)
			return false;
	}
	
	//yes if we got 4 identical not-all-0x00 not-all-0xff replies to "read id"
	return true;
}

bool __attribute__((noinline)) spiRamInit(void)
{
	uint8_t id[MAX_NUM_CHIPS * SIZEOF_ID] = {}, szs[MAX_NUM_CHIPS] = {};
	uint_fast8_t i, j, x, perChipRam, minSzx1, minSzx2, minSzx4;
	const MemWrite32Bfunc *writeFuncs;
	const MemRead32Bfunc *readFuncs;
	
	//first we need to sort out whether we have a a v1.0 - v1.2 board (using 4xSERCOM SPI)
	// or a v1.3+ board (using BBQSPI)
	//we do this by probing for SPI SRAM #0, cause if it is not there, we cannot boot anyways
	//note that on v1.3 board being resetted, the SRAM could be in QSPI mode, thus we check for
	// v1.0-1.2 SRAM config since those are never in QSPI mode. If we do not find it, we assume
	// it is a v1.3+
	if (spiRamPrvIsPreV13board(id))
		spiRamPrvInitV12(id, &readFuncs, &writeFuncs);
	else
		spiRamPrvInitV13(id, &readFuncs, &writeFuncs);
	
	//let's see
	
	pr("RAMs:\n");
	for (i = 0; i < MAX_NUM_CHIPS; i++) {
		
		const char *manuf = 0;
		uint8_t szBase;
		
		pr("   ");
		for (j = 0; j < SIZEOF_ID; j++)
			pr("%02x ", id[i + MAX_NUM_CHIPS * j]);
		
		if (id[i + MAX_NUM_CHIPS * 1]) switch (id[i + MAX_NUM_CHIPS * 0]) {
			case 0x0d:
				manuf = "AP Memory";
				szBase = 2;
				break;
			
			case 0x9d:
				manuf = "ISSI";
				szBase = 1;
				break;
		}
		
		if (!manuf)
			pr("(unrecognized device");
		else {
			uint_fast16_t sz = szBase << (id[i + MAX_NUM_CHIPS * 2] >> 5);
			
			pr(" (%uMB by '%s'", sz, manuf);
			
			if (sz > MAX_CHIP_SIZE)
				pr(" - too big");
			else if (sz < MIN_CHIP_SIZE)
				pr(" - too small");
			else
				szs[i] = sz;
		}
		pr(")\n", szs[i]);
	}
		
	//I used to support multiple ram configs, but it was a pain in my ass in v1.2 board, so i do not anymore
	//v1.3 board does support this since it is simpler in the BBQSPI config
	
	
	//we prefer more ram to faster ram, thus the logic here
	
	minSzx1 = szs[0];
	minSzx2 = szs[0] > szs[1] ? szs[1] : szs[0];
	minSzx4 = szs[2] > szs[3] ? szs[3] : szs[2];
	if (minSzx4 > minSzx2)
		minSzx4 = minSzx2;
	
	if (!minSzx1) {
		
		pr("Cannot run with no RAM0 populated\n");
		return false;
	}
	else if (minSzx1 > minSzx2 * 2 && minSzx1 > minSzx4 * 4) {				//x1 config
		
		perChipRam = minSzx1;
		mWrite32B = writeFuncs[0];
		mRead32B = readFuncs[0];
		x = 1;
	}
	else if (minSzx2 > minSzx4 * 4) {		//x2 config
		
		perChipRam = minSzx2;
		mWrite32B = writeFuncs[1];
		mRead32B = readFuncs[1];
		x = 2;
	}
	else {								//x4 config
		
		perChipRam = minSzx4;
		mWrite32B = writeFuncs[2];
		mRead32B = readFuncs[2];
		x = 4;
	}
	
	mTotalRam = perChipRam * x;
	
	pr("Will run with RAM in %uMBx%u = %uMB config\n", perChipRam, x, mTotalRam);
	
	if (0) {
		static uint8_t __attribute__((aligned(4))) wri[32] = {0xa5,0x5a,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32}, rd[32];
		uint32_t addr = 0x123460;
		uint_fast8_t i;
		
		pr("*** issuing write\n");
		mWrite32B(addr, wri);
		pr("*** issuing read\n");
		mRead32B(addr, rd);
		pr("*** done");
		
		for (i = 0; i < 32; i++) {
			pr(" [%u] %02x -> %0x\n", i, wri[i], rd[i]);
		}
		
		while (1);
	}
	
	spiRamCacheInit();
	return true;
}

uint32_t spiRamGetAmt(void)
{
	return ((uint32_t)mTotalRam) << 20;
}
